import logging
import math

from django import http

from . import version

MAKE_PUBLIC_CODENAME = "make_view_public"


def can_make_public(user):
    return user.has_perm(f"data_browser.{MAKE_PUBLIC_CODENAME}")


def JsonResponse(data):
    res = http.JsonResponse(data, safe=False)
    res["X-Version"] = version
    res["Access-Control-Expose-Headers"] = "X-Version"
    return res


def HttpResponse(*args, **kwargs):
    res = http.HttpResponse(*args, **kwargs)
    res["X-Version"] = version
    res["Access-Control-Expose-Headers"] = "X-Version"
    return res


def debug_log(msg):  # pragma: no cover
    if settings.DEBUG:
        logging.getLogger(__name__).warning(f"DDB: {msg}")


def all_subclasses(cls):
    res = set()
    queue = {cls}
    while queue:
        cls = queue.pop()
        subs = set(cls.__subclasses__())
        queue.update(subs - res)
        res.update(subs)
    return res


def get_optimal_decimal_places(nums, sf=3, max_dp=6):
    actual_dps = set()
    filtered = set()
    for num in nums:
        if num:
            s = f"{num:g}"
            filtered.add(num)
            if "e-" in s:
                actual_dps.add(float("inf"))
            elif "." in s:
                actual_dps.add(len(s.split(".")[1]))
            else:
                actual_dps.add(0)

    if not filtered:
        return 0

    max_actual_dp = max(actual_dps)
    if max_actual_dp <= 2:
        return max_actual_dp

    min_value = min(filtered)
    min_magnitude = math.floor(math.log(min_value, 10))
    dp_for_sf = sf - min_magnitude - 1

    return max(0, min(dp_for_sf, max_actual_dp, max_dp))


class Settings:
    _defaults = {
        "DATA_BROWSER_ALLOW_PUBLIC": False,
        "DATA_BROWSER_AUTH_USER_COMPAT": True,
        "DATA_BROWSER_DEFAULT_ROW_LIMIT": 1000,
        "DATA_BROWSER_DEV": False,
        "DATA_BROWSER_FE_DSN": None,
    }

    def __getattr__(self, name):
        from django.conf import settings

        if hasattr(settings, name):
            return getattr(settings, name)
        return self._defaults[name]


settings = Settings()



import json

import django.contrib.admin.views.decorators as admin_decorators
from django.shortcuts import get_object_or_404

from .common import HttpResponse, JsonResponse, can_make_public
from .models import View, global_data


def deserialize(request):
    data = json.loads(request.body)
    if "model" in data:
        data["model_name"] = data.pop("model")

    res = {
        f: data[f]
        for f in [
            "name",
            "description",
            "public",
            "model_name",
            "fields",
            "query",
            "limit",
        ]
        if f in data
    }

    if "limit" in res:
        try:
            res["limit"] = int(res["limit"])
        except:  # noqa: E722  input sanitization
            res["limit"] = 1
        if res["limit"] < 1:
            res["limit"] = 1

    if not can_make_public(request.user):
        res["public"] = False

    return res


def serialize(view):
    return {
        "name": view.name,
        "description": view.description,
        "public": view.public,
        "model": view.model_name,
        "fields": view.fields,
        "query": view.query,
        "limit": view.limit,
        "publicLink": view.public_link(),
        "googleSheetsFormula": view.google_sheets_formula(),
        "link": f"/query/{view.model_name}/{view.fields}.html?{view.query}&limit={view.limit}",
        "createdTime": f"{view.created_time:%Y-%m-%d %H:%M:%S}",
        "pk": view.pk,
    }


def get_queryset(request):
    return View.objects.filter(owner=request.user)


@admin_decorators.staff_member_required
def view_list(request):
    global_data.request = request

    if request.method == "GET":
        return JsonResponse(
            [
                serialize(view)
                for view in get_queryset(request).order_by("name", "created_time")
            ]
        )
    elif request.method == "POST":
        view = View.objects.create(owner=request.user, **deserialize(request))
        return JsonResponse(serialize(view))
    else:
        return HttpResponse(status=400)


@admin_decorators.staff_member_required
def view_detail(request, pk):
    global_data.request = request
    view = get_object_or_404(get_queryset(request), pk=pk)

    if request.method == "GET":
        return JsonResponse(serialize(view))
    elif request.method == "PATCH":
        for k, v in deserialize(request).items():
            setattr(view, k, v)
        view.save()
        return JsonResponse(serialize(view))
    elif request.method == "DELETE":
        view.delete()
        return HttpResponse(status=204)
    else:
        return HttpResponse(status=400)



from django.db import models

from .common import debug_log
from .types import (
    BooleanType,
    DateTimeType,
    DateType,
    DurationType,
    JSONType,
    NumberArrayType,
    NumberChoiceArrayType,
    NumberChoiceType,
    NumberType,
    StringArrayType,
    StringChoiceArrayType,
    StringChoiceType,
    StringType,
    UnknownType,
)

try:
    from django.contrib.postgres.fields import ArrayField
except ModuleNotFoundError:  # pragma: postgres
    ArrayField = None.__class__

try:
    from django.db.models import JSONField
except ImportError:  # pragma: django < 3.1
    try:
        from django.contrib.postgres.fields import JSONField
    except ModuleNotFoundError:  # pragma: postgres
        JSONField = None.__class__

_STRING_FIELDS = (
    models.CharField,
    models.TextField,
    models.GenericIPAddressField,
    models.UUIDField,
)
_NUMBER_FIELDS = (
    models.DecimalField,
    models.FloatField,
    models.IntegerField,
    models.AutoField,
)
_FIELD_TYPE_MAP = {
    models.BooleanField: BooleanType,
    models.DurationField: DurationType,
    models.NullBooleanField: BooleanType,
    models.DateTimeField: DateTimeType,
    models.DateField: DateType,
    **{f: StringType for f in _STRING_FIELDS},
    **{f: NumberType for f in _NUMBER_FIELDS},
}


def _fmt_choices(choices):
    return [(value, str(label)) for value, label in choices or []]


def get_field_type(field):
    if isinstance(field, ArrayField) and isinstance(
        field.base_field, _STRING_FIELDS
    ):  # pragma: postgres
        base_field, choices = get_field_type(field.base_field)
        array_types = {
            StringType: StringArrayType,
            NumberType: NumberArrayType,
            StringChoiceType: StringChoiceArrayType,
            NumberChoiceType: NumberChoiceArrayType,
        }
        if base_field in array_types:
            return array_types[base_field], choices
        else:
            debug_log(
                f"{field.model.__name__}.{field.name} unsupported subarray type {type(field.base_field).__name__}"
            )
            return UnknownType, None

    elif isinstance(field, ArrayField) and isinstance(
        field.base_field, _NUMBER_FIELDS
    ):  # pragma: postgres
        if field.base_field.choices:
            return NumberChoiceArrayType, _fmt_choices(field.base_field.choices)
        else:
            return NumberArrayType, None
    elif isinstance(field, JSONField):
        res = JSONType
    elif field.__class__ in _FIELD_TYPE_MAP:
        res = _FIELD_TYPE_MAP[field.__class__]
    else:
        for django_type, field_type in _FIELD_TYPE_MAP.items():
            if isinstance(field, django_type):
                res = field_type
                break
        else:
            debug_log(
                f"{field.model.__name__}.{field.name} unsupported type {type(field).__name__}"
            )
            res = UnknownType

    # Choice fields have different lookups
    if res is StringType and field.choices:
        return StringChoiceType, _fmt_choices(field.choices)
    elif res is NumberType and field.choices:
        return NumberChoiceType, _fmt_choices(field.choices)
    else:
        return res, None



import csv
import json
from datetime import datetime

import django
import pytest
from django.contrib.auth.models import User
from django.utils import timezone

import data_browser.models

from .core import models
from .util import update_fe_fixture


def dump(val):
    print(json.dumps(val, indent=4, sort_keys=True))


@pytest.fixture
def products(db):
    address = models.Address.objects.create(city="london")
    producer = models.Producer.objects.create(name="Bob", address=address)
    models.Product.objects.create(name="a", size=1, size_unit="g", producer=producer)
    models.Product.objects.create(name="b", size=1, size_unit="g", producer=producer)
    models.Product.objects.create(name="c", size=2, size_unit="g", producer=producer)


@pytest.fixture
def pivot_products(db):
    address = models.Address.objects.create(city="london", street="bad")
    producer = models.Producer.objects.create(name="Bob", address=address)
    datetimes = [
        datetime(2020, 1, 1, tzinfo=timezone.utc),
        datetime(2020, 2, 1, tzinfo=timezone.utc),
        datetime(2020, 2, 2, tzinfo=timezone.utc),
        datetime(2021, 1, 1, tzinfo=timezone.utc),
        datetime(2021, 1, 2, tzinfo=timezone.utc),
        datetime(2021, 1, 3, tzinfo=timezone.utc),
    ]
    for i, dt in enumerate(datetimes):
        models.Product.objects.create(
            created_time=dt, name=str(dt), size=i + 1, producer=producer
        )


@pytest.mark.skipif(django.VERSION < (2, 2), reason="Django version 2.2 required")
def test_query_html(admin_client, snapshot):
    res = admin_client.get(
        "/data_browser/query/core.Product/size-0,name+1,size_unit.html?size__lt=2&id__gt=0"
    )
    assert res.status_code == 200
    config = json.loads(res.context["config"])
    snapshot.assert_match(config, "config")


def test_query_query(admin_client, snapshot):
    res = admin_client.get(
        "/data_browser/query/core.Product/size-0,name+1,size_unit.query?size__lt=2&id__gt=0"
    )
    assert res.status_code == 200
    query = json.loads(res.content.decode("utf-8"))
    snapshot.assert_match(query, "query")


@pytest.mark.parametrize("format", ["sql", "profile", "pstats", "profile_sql"])
def test_query_misc_formats(admin_client, format):
    # we're not going to check the result as they vary and it's sufficient that it doesn't blow up
    res = admin_client.get(
        f"/data_browser/query/core.Product/size-0,name+1,size_unit.{format}?size__lt=2&id__gt=0"
    )
    assert res.status_code == 200


@pytest.mark.skipif(django.VERSION < (2, 1), reason="Django version 2.1 required")
def test_query_explain(admin_client):
    res = admin_client.get(
        f"/data_browser/query/core.Product/size-0,name+1,size_unit.explain?size__lt=2&id__gt=0"
    )
    assert res.status_code == 200


def test_query_sql_aggregate(admin_client):
    res = admin_client.get(f"/data_browser/query/core.Product/size__count.sql")
    assert res.status_code == 200


@pytest.mark.parametrize(
    "format", ["bad", "profile_bad", "pstats_bad", "profilesql", "pstatsbad"]
)
def test_query_bad_formats(admin_client, format):
    res = admin_client.get(
        f"/data_browser/query/core.Product/size-0,name+1,size_unit.{format}?size__lt=2&id__gt=0"
    )
    assert res.status_code == 404


@pytest.mark.skipif(django.VERSION < (2, 2), reason="Django version 2.2 required")
def test_query_html_no_perms(admin_user, admin_client, snapshot):
    admin_user.is_superuser = False
    admin_user.save()
    res = admin_client.get("/data_browser/query//.html?")
    assert res.status_code == 200
    config = json.loads(res.context["config"])
    snapshot.assert_match(config, "config")


@pytest.mark.skipif(django.VERSION < (2, 2), reason="Django version 2.2 required")
def test_query_ctx(admin_client, snapshot):
    res = admin_client.get("/data_browser/query//.ctx?")
    assert res.status_code == 200
    config = res.json()
    snapshot.assert_match(config, "config")
    update_fe_fixture("frontend/src/context_fixture.json", config)


@pytest.mark.usefixtures("products")
def test_query_json_bad_fields(admin_client):
    res = admin_client.get(
        "".join(
            [
                "/data_browser/query/core.Product/",
                "size-0,name+1,size_unit,bob-2,is_onsale,pooducer__name,producer__name.json",
                "?size__lt=2&id__gt=0&bob__gt=1&size__xx=1&size__lt=xx",
            ]
        )
    )
    assert res.status_code == 200
    assert json.loads(res.content.decode("utf-8"))["rows"] == [
        {
            "size": 1,
            "name": "a",
            "size_unit": "g",
            "is_onsale": "False",
            "producer__name": "Bob",
        },
        {
            "size": 1,
            "name": "b",
            "size_unit": "g",
            "is_onsale": "False",
            "producer__name": "Bob",
        },
    ]


def test_query_bad_media(admin_client):
    res = admin_client.get(
        "/data_browser/query/core.Product/size-0,name+1,size_unit.bob?size__lt=2&id__gt=0"
    )
    assert res.status_code == 404


@pytest.mark.usefixtures("products")
def test_query_csv(admin_client):
    res = admin_client.get(
        "/data_browser/query/core.Product/size-0,name+1,size_unit.csv?size__lt=2&id__gt=0"
    )
    assert res.status_code == 200
    res = res.getvalue().decode("utf-8")
    dump(res)
    rows = list(csv.reader(res.splitlines()))
    dump(rows)
    assert rows == [["Size", "Name", "Size unit"], ["1.0", "a", "g"], ["1.0", "b", "g"]]


@pytest.mark.usefixtures("pivot_products")
def test_query_csv_pivoted(admin_client):
    res = admin_client.get(
        "/data_browser/query/core.Product/created_time__year+0,&created_time__month+1,id__count,size__max.csv?"
    )
    assert res.status_code == 200
    res = res.getvalue().decode("utf-8")
    dump(res)
    rows = list(csv.reader(res.splitlines()))
    dump(rows)
    assert rows == [
        ["Created time month", "January", "", "Feburary", ""],
        ["Created time year", "ID count", "Size max", "ID count", "Size max"],
        ["2020.0", "1.0", "1.0", "2.0", "3.0"],
        ["2021.0", "3.0", "6.0", "", ""],
    ]


testdata = [
    ("----"),
    ("---b"),
    ("--c-"),
    ("--cb"),
    ("-r--"),
    ("-r-b"),
    ("-rc-"),
    ("-rcb"),
    ("d---"),
    ("d--b"),
    ("d-c-"),
    ("d-cb"),
    ("dr--"),
    ("dr-b"),
    ("drc-"),
    ("drcb"),
]


@pytest.mark.usefixtures("pivot_products")
@pytest.mark.parametrize("key", testdata)
def test_query_csv_pivot_permutations(admin_client, key, snapshot):
    fields = []
    if "r" in key:
        fields.append("created_time__year+0")
    if "c" in key:
        fields.append("&created_time__month+1")
    if "b" in key:
        fields.extend(["id__count", "size__max"])
    filters = "" if "d" in key else "id__equals=-1"

    res = admin_client.get(
        f"/data_browser/query/core.Product/{','.join(fields)}.csv?{filters}"
    )
    assert res.status_code == 200
    res = res.getvalue().decode("utf-8")
    dump(res)
    rows = list(csv.reader(res.splitlines()))
    dump(rows)
    snapshot.assert_match(rows, "key")


@pytest.mark.usefixtures("products")
def test_query_json(admin_client, snapshot):
    res = admin_client.get(
        "/data_browser/query/core.Product/size-0,name+1,size_unit.json?size__lt=2&id__gt=0"
    )
    assert res.status_code == 200
    data = json.loads(res.content.decode("utf-8"))
    snapshot.assert_match(data, "data")


@pytest.mark.usefixtures("pivot_products")
def test_query_json_pivot(admin_client, snapshot):
    res = admin_client.get(
        "/data_browser/query/core.Product/created_time__year+0,&created_time__month+1,id__count,size__max.json?"
    )
    assert res.status_code == 200
    data = json.loads(res.content.decode("utf-8"))
    snapshot.assert_match(data, "data")


@pytest.mark.usefixtures("products")
def test_query_json_bad_model(admin_client):
    res = admin_client.get(
        "/data_browser/query/core.Bob/size-0,name+1,size_unit.json?size__lt=2&id__gt=0"
    )
    assert res.status_code == 404


@pytest.mark.usefixtures("products")
def test_view_csv(admin_client, settings):
    view = data_browser.models.View.objects.create(
        model_name="core.Product",
        fields="size-0,name+1,size_unit",
        query="size__lt=2&id__gt=0",
        owner=User.objects.get(),
    )

    res = admin_client.get(f"/data_browser/view/{view.public_slug}.csv")
    assert res.status_code == 404

    view.public = True
    view.save()
    res = admin_client.get(f"/data_browser/view/{view.public_slug}.csv")
    assert res.status_code == 200
    res = res.getvalue().decode("utf-8")
    dump(res)
    rows = list(csv.reader(res.splitlines()))
    dump(rows)
    assert rows == [["Size", "Name", "Size unit"], ["1.0", "a", "g"], ["1.0", "b", "g"]]

    settings.DATA_BROWSER_ALLOW_PUBLIC = False
    res = admin_client.get(f"/data_browser/view/{view.public_slug}.csv")
    assert res.status_code == 404
    settings.DATA_BROWSER_ALLOW_PUBLIC = True

    view.owner = User.objects.create(is_staff=True)
    view.save()
    res = admin_client.get(f"/data_browser/view/{view.public_slug}.csv")
    assert res.status_code == 404


@pytest.mark.usefixtures("products")
def test_view_json(admin_client):
    view = data_browser.models.View.objects.create(
        model_name="core.Product",
        fields="size-0,name+1,size_unit",
        query="size__lt=2&id__gt=0",
        owner=User.objects.get(),
    )

    res = admin_client.get(f"/data_browser/view/{view.public_slug}.json")
    assert res.status_code == 404

    view.public = True
    view.save()
    res = admin_client.get(f"/data_browser/view/{view.public_slug}.json")
    assert res.status_code == 200
    data = json.loads(res.content.decode("utf-8"))
    dump(data)
    assert data == {
        "rows": [
            {"size": 1, "name": "a", "size_unit": "g"},
            {"size": 1, "name": "b", "size_unit": "g"},
        ],
        "cols": [{}],
        "body": [[{}, {}]],
        "length": 2,
        "formatHints": {
            "name": {},
            "size": {
                "highCutOff": 10000000000.0,
                "lowCutOff": 0.0001,
                "maximumFractionDigits": 0,
                "minimumFractionDigits": 0,
                "significantFigures": 3,
            },
            "size_unit": {},
        },
    }

    view.owner = User.objects.create(is_staff=True)
    view.save()
    res = admin_client.get(f"/data_browser/view/{view.public_slug}.csv")
    assert res.status_code == 404



from django.contrib import admin

from .core.admin import AddressAdmin, ProductAdmin
from .core.models import Address, Producer, Product


class TestAdminMixin:
    def test_annotation_loaded_on_detail(self, admin_client, mocker):
        get_queryset = mocker.patch(
            "tests.core.admin.AddressAdmin.andrew.get_queryset",
            wraps=AddressAdmin.andrew.get_queryset,
        )
        a = Address.objects.create()
        assert (
            admin_client.get(f"/admin/core/address/{a.pk}/change/").status_code == 200
        )
        get_queryset.assert_called_once()

    def test_annotation_loaded_unnecessarily_on_detail(self, admin_client, mocker):
        get_queryset = mocker.patch(
            "tests.core.admin.ProductAdmin.annotated.get_queryset",
            wraps=ProductAdmin.annotated.get_queryset,
        )
        p = Product.objects.create(producer=Producer.objects.create())
        assert (
            admin_client.get(f"/admin/core/product/{p.pk}/change/").status_code == 200
        )
        get_queryset.assert_called_once()

    def test_annotation_loaded_on_list(self, admin_client, mocker):
        get_queryset = mocker.patch(
            "tests.core.admin.ProductAdmin.annotated.get_queryset",
            wraps=ProductAdmin.annotated.get_queryset,
        )
        assert admin_client.get(f"/admin/core/product/").status_code == 200
        get_queryset.assert_called_once()

    def test_annotation_not_loaded_on_list(self, admin_client, mocker):
        get_queryset = mocker.patch(
            "tests.core.admin.AddressAdmin.andrew.get_queryset",
            wraps=AddressAdmin.andrew.get_queryset,
        )
        assert admin_client.get(f"/admin/core/address/").status_code == 200
        get_queryset.assert_not_called()

    def test_request_factory_compability_list(self, rf, admin_user, mocker):
        request = rf.get("/")
        request.user = admin_user
        get_queryset = mocker.patch(
            "tests.core.admin.AddressAdmin.andrew.get_queryset",
            wraps=AddressAdmin.andrew.get_queryset,
        )
        resp = AddressAdmin(Address, admin.site).changelist_view(request)
        assert resp.status_code == 200
        get_queryset.assert_called_once()

    def test_request_factory_compability_detail(self, rf, admin_user, mocker):
        address = Address.objects.create()
        request = rf.get("/")
        request.user = admin_user
        get_queryset = mocker.patch(
            "tests.core.admin.AddressAdmin.andrew.get_queryset",
            wraps=AddressAdmin.andrew.get_queryset,
        )
        resp = AddressAdmin(Address, admin.site).changeform_view(
            request, str(address.pk)
        )
        assert resp.status_code == 200
        get_queryset.assert_called_once()

    def test_ddb_url(self, admin_client):
        resp = admin_client.get(f"/admin/core/product/")
        assert resp.status_code == 200
        assert (
            resp.context["ddb_url"]
            == "/data_browser/query/core.Product/.html?a_field__a_lookup=a_value&name__not_equals=not+a+thing&a_field__a_lookup=true"
        )

    def test_ddb_url_ignored(self, admin_client):
        resp = admin_client.get(f"/admin/core/ignored/")
        assert resp.status_code == 200
        assert "ddb_url" not in resp.context



import dj_database_url
import django
import pytest
from django.conf import settings

DATABASE_CONFIG = dj_database_url.config(
    conn_max_age=600, default="sqlite:///db.sqlite3"
)

POSTGRES = "postgresql" in DATABASE_CONFIG["ENGINE"]
SQLITE = "sqlite" in DATABASE_CONFIG["ENGINE"]

if POSTGRES:  # pragma: postgres
    JSON_FIELD_SUPPORT = django.VERSION >= (2, 1)
    ARRAY_FIELD_SUPPORT = True
else:  # pragma: not postgres
    JSON_FIELD_SUPPORT = django.VERSION >= (3, 1)
    ARRAY_FIELD_SUPPORT = False

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django.contrib.admin",
    "tests.core",
    "data_browser",
]

if JSON_FIELD_SUPPORT:  # pragma: no branch
    INSTALLED_APPS.append("tests.json")
if ARRAY_FIELD_SUPPORT:  # pragma: postgres
    INSTALLED_APPS.append("tests.array")

settings.configure(
    INSTALLED_APPS=INSTALLED_APPS,
    DATABASES={"default": DATABASE_CONFIG},
    ROOT_URLCONF="tests.urls",
    MIDDLEWARE=[
        "django.middleware.security.SecurityMiddleware",
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.middleware.common.CommonMiddleware",
        "django.middleware.csrf.CsrfViewMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
    ],
    TEMPLATES=[
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "OPTIONS": {
                "context_processors": [
                    "django.template.context_processors.debug",
                    "django.template.context_processors.request",
                    "django.template.context_processors.static",
                    "django.contrib.auth.context_processors.auth",
                ],
                "loaders": ["django.template.loaders.app_directories.Loader"],
            },
        }
    ],
    STATIC_URL="/static/",
    MEDIA_URL="/media/",
    DATA_BROWSER_ALLOW_PUBLIC=True,
    USE_I18N=True,
    USE_TZ=True,
    TIME_ZONE="UTC",
)


@pytest.fixture
def req(rf, admin_user):
    req = rf.get("/")
    req.user = admin_user
    return req



import os

from django.urls import path, re_path, register_converter
from django.urls.converters import StringConverter
from django.views.generic.base import RedirectView
from django.views.static import serve

from .api import view_detail, view_list
from .common import settings
from .views import proxy_js_dev_server, query, query_ctx, query_html, view

FE_BUILD_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "fe_build")
WEB_ROOT_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "web_root")


class OptionalString(StringConverter):
    regex = "[^/]*"


register_converter(OptionalString, "opt_str")


app_name = "data_browser"

QUERY_PATH = "query/<model_name>/<opt_str:fields>"

if settings.DATA_BROWSER_DEV:  # pragma: no cover
    static_view = (proxy_js_dev_server,)
else:
    static_view = (serve, {"document_root": FE_BUILD_DIR})

urlpatterns = [
    # queries
    path(f"{QUERY_PATH}.html", query_html, name="query_html"),
    path(f"{QUERY_PATH}.ctx", query_ctx),
    path(f"{QUERY_PATH}.<media>", query, name="query"),
    # views
    path("view/<pk>.<media>", view, name="view"),
    # api
    path("api/views/", view_list, name="view_list"),
    path("api/views/<pk>/", view_detail, name="view_detail"),
    # other html pages
    re_path(r".*\.html", query_html),
    re_path(r".*\.ctx", query_ctx),
    path("", query_html, name="home"),
    # static files
    re_path(r"^(?P<path>static/.*)$", *static_view, name="static"),
    re_path(
        r"^.*/(?P<path>static/.*)$",
        RedirectView.as_view(pattern_name="data_browser:static", permanent=True),
    ),
    re_path(r"^(?P<path>.*)$", serve, {"document_root": WEB_ROOT_DIR}),
]



import json
import re
from functools import lru_cache

import dateutil.parser
from django.conf import settings
from django.utils import dateparse, html, timezone

from .common import all_subclasses, get_optimal_decimal_places

ASC, DSC = "asc", "dsc"


class TypeMeta(type):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.raw_type and self.element_type:
            assert self.raw_type.element_type == self.element_type.raw_type

    def __repr__(cls):
        return cls.__name__

    @property
    def default_lookup(cls):
        lookups = cls.lookups
        return list(lookups)[0] if lookups else None

    @property
    def lookups(cls):
        return {name: type_.name for name, type_ in cls._lookups().items()}

    @property
    def name(cls):
        name = cls.__name__.lower()
        assert name.endswith("type")
        return name[: -len("type")]


class BaseType(metaclass=TypeMeta):
    default_value = None
    choices = ()
    raw_type = None
    element_type = None

    def __init__(self):
        assert False

    @staticmethod
    def _lookups():
        return {}

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: value

    @classmethod
    def get_formatter(cls, choices):
        return cls._get_formatter(choices)

    @staticmethod
    def _parse(value, choices):
        return value

    @classmethod
    def parse(cls, lookup, value, choices):
        lookups = cls.lookups
        if lookup not in lookups:
            return None, f"Bad lookup '{lookup}' expected {lookups}"
        else:
            type_ = TYPES[lookups[lookup]]
            try:
                return type_._parse(value, choices), None
            except Exception as e:
                err_message = str(e) if str(e) else repr(e)
                return None, err_message

    @staticmethod
    def get_format_hints(name, data):
        return {}


class StringType(BaseType):
    default_value = ""

    @classmethod
    def _lookups(cls):
        return {
            "equals": cls,
            "contains": cls,
            "starts_with": cls,
            "ends_with": cls,
            "regex": RegexType,
            "not_equals": cls,
            "not_contains": cls,
            "not_starts_with": cls,
            "not_ends_with": cls,
            "not_regex": RegexType,
            "is_null": IsNullType,
        }


class NumberType(BaseType):
    default_value = 0

    @classmethod
    def _lookups(cls):
        return {
            "equals": cls,
            "not_equals": cls,
            "gt": cls,
            "gte": cls,
            "lt": cls,
            "lte": cls,
            "is_null": IsNullType,
        }

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: None if value is None else float(value)

    @staticmethod
    def _parse(value, choices):
        return float(value)

    @staticmethod
    def get_format_hints(name, data):
        nums = [
            row[name] for row in data if row and row[name] and abs(row[name] > 0.0001)
        ]
        dp = get_optimal_decimal_places(nums)
        return {
            "minimumFractionDigits": dp,
            "maximumFractionDigits": dp,
            "significantFigures": 3,
            "lowCutOff": 0.0001,
            "highCutOff": 1e10,
        }


class RegexType(BaseType):
    default_value = ".*"

    @staticmethod
    @lru_cache(maxsize=None)
    def _parse(value, choices):
        from django.contrib.contenttypes.models import ContentType
        from django.db.transaction import atomic

        # this is dirty
        # we need to check if the regex is going to cause a db exception
        # and not kill any in progress transaction as we check
        with atomic():
            list(ContentType.objects.filter(model__regex=value))
        return value


class DurationType(BaseType):
    default_value = ""

    @classmethod
    def _lookups(cls):
        return {
            "equals": cls,
            "not_equals": cls,
            "gt": cls,
            "gte": cls,
            "lt": cls,
            "lte": cls,
            "is_null": IsNullType,
        }

    @staticmethod
    def _parse(value, choices):
        if value.count(":") == 1:
            value += ":0"

        res = dateparse.parse_duration(value)
        assert res is not None, "Duration value should be 'DD HH:MM:SS'"
        return res

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: None if value is None else str(value)


class DateTypeMixin:
    @classmethod
    def _lookups(cls):
        return {
            "equals": cls,
            "not_equals": cls,
            "gt": cls,
            "gte": cls,
            "lt": cls,
            "lte": cls,
            "is_null": IsNullType,
        }

    @classmethod
    def _parse(cls, value, choices):
        d8 = r"(\d{8})"
        d422 = r"(\d{4}[^\d]*\d{2}[^\d]*\d{2})"
        if re.match(r"[^\d]*(" + d8 + "|" + d422 + ")", value):
            # looks like some kinda iso date, roll with the defaults
            res = dateutil.parser.parse(value)
        else:
            res = {
                dateutil.parser.parse(value, dayfirst=False, yearfirst=False),
                dateutil.parser.parse(value, dayfirst=True, yearfirst=False),
                dateutil.parser.parse(value, dayfirst=False, yearfirst=True),
                dateutil.parser.parse(value, dayfirst=True, yearfirst=True),
            }
            assert len(res) == 1, "Ambiguous value"
            res = res.pop()
        return timezone.make_aware(res)


class DateTimeType(DateTypeMixin, BaseType):
    default_value = "now"

    @classmethod
    def _parse(cls, value, choices):
        if value.lower().strip() == "now":
            return timezone.now()
        return super()._parse(value, choices)

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        if settings.USE_TZ:
            return (
                lambda value: None if value is None else str(timezone.make_naive(value))
            )
        else:
            return lambda value: None if value is None else str(value)


class DateType(DateTypeMixin, BaseType):
    default_value = "today"

    @classmethod
    def _parse(cls, value, choices):
        if value.lower().strip() == "today":
            return timezone.now().date()
        return super()._parse(value, choices).date()

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: None if value is None else str(value)


class HTMLType(StringType):
    @staticmethod
    def _lookups():
        return {}

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: None if value is None else html.conditional_escape(value)


class BooleanType(BaseType):
    default_value = True

    @classmethod
    def _lookups(cls):
        return {"equals": cls, "not_equals": cls, "is_null": IsNullType}

    @staticmethod
    def _parse(value, choices):
        value = value.lower()
        if value == "true":
            return True
        elif value == "false":
            return False
        else:
            raise ValueError("Expected 'true' or 'false'")

    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: None if value is None else bool(value)


class UnknownType(BaseType):
    @staticmethod
    def _get_formatter(choices):
        assert not choices
        return lambda value: None if value is None else str(value)

    @staticmethod
    def _lookups():
        return {"is_null": IsNullType}


def _json_loads(value):
    try:
        return json.loads(value.strip())
    except json.JSONDecodeError:
        raise ValueError("Invalid JSON value")


class JSONFieldType(BaseType):
    default_value = "|"

    @staticmethod
    def _parse(value, choices):
        value = value.strip()
        if "|" not in value:
            raise ValueError("Missing seperator '|'")
        field, value = value.split("|", 1)
        if not field:
            raise ValueError("Invalid field name")
        return [field, _json_loads(value)]


class JSONType(BaseType):
    @classmethod
    def _lookups(cls):
        return {
            "equals": cls,
            "has_key": StringType,
            "field_equals": JSONFieldType,
            "not_equals": cls,
            "not_has_key": StringType,
            "not_field_equals": JSONFieldType,
            "is_null": IsNullType,
        }

    @staticmethod
    def _parse(value, choices):
        return _json_loads(value)


class ChoiceTypeMixin:
    default_value = None

    @classmethod
    def _get_formatter(cls, choices):
        assert choices
        choices = dict(choices)
        return lambda value: choices.get(value, value)

    @classmethod
    def _parse(cls, value, choices):
        assert choices
        choices = {v: k for k, v in choices}
        if value not in choices:
            raise ValueError(f"Unknown choice '{value}'")
        return choices[value]

    @classmethod
    def _lookups(cls):
        return {"equals": cls, "not_equals": cls, "is_null": IsNullType}


class StringChoiceType(ChoiceTypeMixin, BaseType):
    raw_type = StringType


class NumberChoiceType(ChoiceTypeMixin, BaseType):
    raw_type = NumberType


class IsNullType(ChoiceTypeMixin, BaseType):
    choices = [(None, "IsNull"), (True, "IsNull"), (False, "NotNull")]
    default_value = choices[0][1]

    @staticmethod
    def _lookups():
        return {"equals": IsNullType}

    @classmethod
    def _get_formatter(cls, choices):
        return super()._get_formatter(cls.choices)

    @classmethod
    def _parse(cls, value, choices):
        return super()._parse(value, cls.choices)


class ArrayTypeMixin:
    default_value = None

    @classmethod
    def _get_formatter(cls, choices):  # pragma: postgres
        element_formatter = cls.element_type._get_formatter(choices)
        return (
            lambda value: None
            if value is None
            else json.dumps([element_formatter(v) for v in value])
        )

    @classmethod
    def _lookups(cls):
        return {
            "equals": cls,
            "contains": cls.element_type,
            "length": NumberType,
            "not_equals": cls,
            "not_contains": cls.element_type,
            "not_length": NumberType,
            "is_null": IsNullType,
        }

    @classmethod
    def _parse(cls, value, choices):
        value = _json_loads(value)
        if not isinstance(value, list):
            raise ValueError("Expected a list")
        return [cls.element_type._parse(v, choices) for v in value]


class StringArrayType(ArrayTypeMixin, BaseType):
    element_type = StringType


class NumberArrayType(ArrayTypeMixin, BaseType):
    element_type = NumberType


class StringChoiceArrayType(ArrayTypeMixin, BaseType):
    element_type = StringChoiceType
    raw_type = StringArrayType


class NumberChoiceArrayType(ArrayTypeMixin, BaseType):
    element_type = NumberChoiceType
    raw_type = NumberArrayType


TYPES = {cls.name: cls for cls in all_subclasses(BaseType)}



# Generated by Django 2.0.13 on 2020-05-16 16:26

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("data_browser", "0004_auto_20200501_0903"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="view",
            options={
                "permissions": [
                    ("make_view_public", "Can make a saved view publically available")
                ]
            },
        ),
    ]



from dataclasses import dataclass
from typing import Sequence, Tuple

from django.db import models
from django.db.models import OuterRef, Subquery
from django.utils.html import format_html

from .types import ASC, BaseType, BooleanType, DateTimeType, DateType, HTMLType
from .util import s


@dataclass
class OrmBoundField:
    field: "OrmBaseField"
    previous: "OrmBoundField"
    full_path: Sequence[str]
    pretty_path: Sequence[str]
    queryset_path: Sequence[str]
    aggregate_clause: Tuple[str, models.Func] = None
    filter_: bool = False
    having: bool = False
    model_name: str = None

    @property
    def path_str(self):
        return s(self.full_path)

    @property
    def queryset_path_str(self):
        return s(self.queryset_path)

    @property
    def group_by(self):
        return self.field.can_pivot

    def _lineage(self):
        if self.previous:
            return self.previous._lineage() + [self]
        return [self]

    def annotate(self, request, qs):
        for field in self._lineage():
            qs = field._annotate(request, qs)
        return qs

    def _annotate(self, request, qs):
        return qs

    def __getattr__(self, name):
        return getattr(self.field, name)

    def parse(self, lookup, value):
        return self.type_.parse(lookup, value, self.choices)

    @classmethod
    def blank(cls):
        return cls(
            field=None, previous=None, full_path=[], pretty_path=[], queryset_path=[]
        )

    def get_format_hints(self, data):
        hints = self.type_.get_format_hints(self.path_str, data)
        return {**hints, **(self.format_hints or {})}


@dataclass
class OrmBaseField:
    model_name: str
    name: str
    pretty_name: str
    type_: BaseType = None
    concrete: bool = False
    rel_name: str = None
    can_pivot: bool = False
    choices: Sequence[Tuple[str, str]] = ()
    default_sort: str = None
    format_hints: dict = None

    def __post_init__(self):
        if not self.type_:
            assert self.rel_name
        if self.concrete or self.can_pivot:
            assert self.type_

    def get_formatter(self):
        return self.type_.get_formatter(self.choices)


class OrmFkField(OrmBaseField):
    def __init__(self, model_name, name, pretty_name, rel_name):
        super().__init__(model_name, name, pretty_name, rel_name=rel_name)

    def bind(self, previous):
        previous = previous or OrmBoundField.blank()
        return OrmBoundField(
            field=self,
            previous=previous,
            full_path=previous.full_path + [self.name],
            pretty_path=previous.pretty_path + [self.pretty_name],
            queryset_path=previous.queryset_path + [self.name],
        )


class OrmConcreteField(OrmBaseField):
    def __init__(self, model_name, name, pretty_name, type_, rel_name, choices):
        super().__init__(
            model_name,
            name,
            pretty_name,
            concrete=True,
            type_=type_,
            rel_name=rel_name,
            can_pivot=True,
            choices=choices or (),
            default_sort=ASC if type_ in [DateType, DateTimeType] else None,
        )

    def bind(self, previous):
        previous = previous or OrmBoundField.blank()
        return OrmBoundField(
            field=self,
            previous=previous,
            full_path=previous.full_path + [self.name],
            pretty_path=previous.pretty_path + [self.pretty_name],
            queryset_path=previous.queryset_path + [self.name],
            filter_=True,
        )


class OrmRawField(OrmConcreteField):
    def bind(self, previous):
        return OrmBoundField(
            field=self,
            previous=previous,
            full_path=previous.full_path + [self.name],
            pretty_path=previous.pretty_path + [self.pretty_name],
            queryset_path=previous.queryset_path,
            filter_=True,
        )


class OrmCalculatedField(OrmBaseField):
    def __init__(self, model_name, name, pretty_name, func):
        if getattr(func, "boolean", False):
            type_ = BooleanType
        else:
            type_ = HTMLType

        super().__init__(model_name, name, pretty_name, type_=type_, can_pivot=True)
        self.func = func

    def bind(self, previous):
        previous = previous or OrmBoundField.blank()
        return OrmBoundField(
            field=self,
            previous=previous,
            full_path=previous.full_path + [self.name],
            pretty_path=previous.pretty_path + [self.pretty_name],
            queryset_path=previous.queryset_path + ["id"],
            model_name=self.model_name,
        )

    def get_formatter(self):
        base_formatter = super().get_formatter()

        def format(obj):
            if obj is None:
                return None

            try:
                value = self.func(obj)
            except Exception as e:
                return str(e)

            return base_formatter(value)

        return format


class OrmBoundAnnotatedField(OrmBoundField):
    def _annotate(self, request, qs):
        from .orm_results import admin_get_queryset

        return qs.annotate(
            **{
                s(self.queryset_path): Subquery(
                    admin_get_queryset(self.admin, request, [self.name])
                    .filter(pk=OuterRef(s(self.previous.queryset_path + ["id"])))
                    .values(self.name)[:1],
                    output_field=self.django_field,
                )
            }
        )


class OrmAnnotatedField(OrmBaseField):
    def __init__(
        self, model_name, name, pretty_name, type_, django_field, admin, choices
    ):
        super().__init__(
            model_name,
            name,
            pretty_name,
            type_=type_,
            rel_name=type_.name,
            can_pivot=True,
            concrete=True,
            choices=choices or (),
        )
        self.django_field = django_field
        self.admin = admin

    def bind(self, previous):
        previous = previous or OrmBoundField.blank()

        full_path = previous.full_path + [self.name]
        return OrmBoundAnnotatedField(
            field=self,
            previous=previous,
            full_path=full_path,
            pretty_path=previous.pretty_path + [self.pretty_name],
            queryset_path=[s(["ddb"] + full_path)],
            filter_=True,
        )


class OrmFileField(OrmConcreteField):
    def __init__(self, model_name, name, pretty_name, django_field):
        super().__init__(
            model_name,
            name,
            pretty_name,
            type_=HTMLType,
            rel_name=HTMLType.name,
            choices=None,
        )
        self.django_field = django_field

    def get_formatter(self):
        def format(value):
            if not value:
                return None

            try:
                # some storage backends will hard fail if their underlying storage isn't
                # setup right https://github.com/tolomea/django-data-browser/issues/11
                return format_html(
                    '<a href="{}">{}</a>', self.django_field.storage.url(value), value
                )
            except Exception as e:
                return str(e)

        return format



from datetime import date, datetime, timedelta

import django
import pytest
from django.contrib.admin.options import BaseModelAdmin
from django.contrib.auth.models import Permission, User
from django.utils import timezone

from data_browser.orm_admin import OrmModel, get_models
from data_browser.orm_results import admin_get_queryset, get_results
from data_browser.query import BoundQuery, Query

from .core import models
from .util import ANY, KEYS


def sortedAssert(a, b):
    assert sorted(a, key=str) == sorted(b, key=str)


def admin_link(obj):
    model = obj.__class__.__name__.lower()
    return f'<a href="/admin/core/{model}/{obj.pk}/change/">{obj}</a>'


def flatten_table(fields, data):
    return [[(row[f.path_str] if row else None) for f in fields] for row in data]


@pytest.fixture
def products(db):
    now = timezone.now()
    res = []

    address = models.Address.objects.create(city="london", street="bad")
    producer = models.Producer.objects.create(name="Bob", address=address)
    res.append(
        models.Product.objects.create(
            created_time=now,
            name="a",
            size=1,
            size_unit="g",
            producer=producer,
            fake="phoney",
            boat=2.2,
        )
    )

    address = models.Address.objects.create(city="london", street="good")
    producer = models.Producer.objects.create(name="Bob", address=address)
    res.append(
        models.Product.objects.create(
            created_time=now, name="b", size=1, size_unit="g", producer=producer
        )
    )

    producer = models.Producer.objects.create(name="Bob", address=None)
    res.append(
        models.Product.objects.create(
            created_time=now,
            name="c",
            size=2,
            size_unit="g",
            producer=producer,
            boat=2.4,
        )
    )

    return res


@pytest.fixture
def pivot_products(db):
    address = models.Address.objects.create(city="london", street="bad")
    producer = models.Producer.objects.create(name="Bob", address=address)
    datetimes = [
        datetime(2020, 1, 1, tzinfo=timezone.utc),
        datetime(2020, 2, 1, tzinfo=timezone.utc),
        datetime(2020, 2, 2, tzinfo=timezone.utc),
        datetime(2021, 1, 1, tzinfo=timezone.utc),
        datetime(2021, 1, 2, tzinfo=timezone.utc),
        datetime(2021, 1, 3, tzinfo=timezone.utc),
        datetime(2021, 2, 1, tzinfo=timezone.utc),
        datetime(2021, 2, 2, tzinfo=timezone.utc),
        datetime(2021, 2, 3, tzinfo=timezone.utc),
        datetime(2021, 2, 4, tzinfo=timezone.utc),
    ]
    for i, dt in enumerate(datetimes):
        models.Product.objects.create(
            created_time=dt, name=str(dt), size=i + 1, producer=producer
        )


@pytest.fixture
def orm_models(req):
    return get_models(req)


@pytest.fixture
def get_product_flat(req, orm_models, django_assert_num_queries):
    def helper(queries, *args, **kwargs):
        query = Query.from_request("core.Product", *args, **kwargs)
        bound_query = BoundQuery.bind(query, orm_models)
        with django_assert_num_queries(queries):
            data = get_results(req, bound_query, orm_models, False)
            return flatten_table(bound_query.fields, data["rows"])

    return helper


@pytest.fixture
def get_product_pivot(req, orm_models, django_assert_num_queries):
    def helper(queries, *args, **kwargs):
        query = Query.from_request("core.Product", *args, **kwargs)
        bound_query = BoundQuery.bind(query, orm_models)
        with django_assert_num_queries(queries):
            data = get_results(req, bound_query, orm_models, False)
            return {
                "cols": flatten_table(bound_query.col_fields, data["cols"]),
                "rows": flatten_table(bound_query.row_fields, data["rows"]),
                "body": [
                    flatten_table(bound_query.data_fields, row) for row in data["body"]
                ],
            }

    return helper


@pytest.mark.usefixtures("products")
def test_get_results_all(get_product_flat):
    data = get_product_flat(1, "size-0,name+1,size_unit", {})
    assert data == [[2, "c", "g"], [1, "a", "g"], [1, "b", "g"]]


@pytest.mark.usefixtures("products")
def test_get_results_unknown_field(get_product_flat):
    data = get_product_flat(1, "name+!,fake", {})
    assert data == [["a", "phoney"], ["b", "{}"], ["c", "{}"]]


def test_get_admin_link(get_product_flat, products):
    data = get_product_flat(2, "producer__address__admin", {})
    sortedAssert(
        data,
        [
            [None],
            [admin_link(products[0].producer.address)],
            [admin_link(products[1].producer.address)],
        ],
    )


def test_get_file_link(get_product_flat):
    producer = models.Producer.objects.create()
    models.Product.objects.create(name="a", producer=producer)
    models.Product.objects.create(name="b", producer=producer, image="bob.jpg")
    data = get_product_flat(1, "name,image", {})
    sortedAssert(data, [["a", None], ["b", '<a href="/media/bob.jpg">bob.jpg</a>']])


def test_bad_storage(monkeypatch, req):
    # break storage
    from django.core.files.storage import FileSystemStorage, default_storage
    from django.utils.functional import empty

    def boom(*args, **kwargs):
        assert False

    monkeypatch.setattr(FileSystemStorage, "__init__", boom)
    default_storage._wrapped = empty

    # copy what the fixture does
    orm_models = get_models(req)

    def get_product_flat(*args, **kwargs):
        query = Query.from_request("core.Product", *args)
        bound_query = BoundQuery.bind(query, orm_models)
        data = get_results(req, bound_query, orm_models, False)
        return flatten_table(bound_query.fields, data["rows"])

    # some storage backends will hard fail if their underlying storage isn't
    # setup right https://github.com/tolomea/django-data-browser/issues/11
    producer = models.Producer.objects.create()
    models.Product.objects.create(name="a", producer=producer)
    models.Product.objects.create(name="b", producer=producer, image="bob.jpg")
    data = get_product_flat("name,image", {})
    sortedAssert(data, [["a", None], ["b", "assert False"]])


@pytest.mark.usefixtures("products")
def test_get_callable(get_product_flat):
    data = get_product_flat(2, "name,funky,lambda", {})
    sortedAssert(data, [["a", "fa", "la"], ["b", "fb", "lb"], ["c", "fc", "lc"]])


@pytest.mark.usefixtures("products")
def test_get_calculated_field_on_admin(get_product_flat):
    data = get_product_flat(2, "producer__address__bob", {})
    sortedAssert(data, [[None], ["err"], ["bob"]])


@pytest.mark.usefixtures("products")
def test_get_calculated_boolean_field_on_admin(get_product_flat):
    data = get_product_flat(2, "calculated_boolean", {})
    sortedAssert(data, [[True], [True], [True]])


def test_get_annotated_field_at_base(products, get_product_flat, mocker):
    mock = mocker.patch(
        "data_browser.orm_results.admin_get_queryset", wraps=admin_get_queryset
    )
    data = get_product_flat(1, "annotated+1,size-2", {"annotated__not_equals": ["a"]})
    assert data == [["b", 1], ["c", 2]]
    assert len(mock.call_args_list) == 2


@pytest.mark.skipif(django.VERSION < (2, 1), reason="Django version 2.1 required")
def test_get_aggretated_annotated_field_at_base(products, get_product_flat, mocker):
    mock = mocker.patch(
        "data_browser.orm_results.admin_get_queryset", wraps=admin_get_queryset
    )
    data = get_product_flat(1, "annotated__count+1,size-2", {})
    assert data == [[1, 2], [2, 1]]
    assert len(mock.call_args_list) == 2


def test_get_annotated_field_function_at_base(products, get_product_flat, mocker):
    mock = mocker.patch(
        "data_browser.orm_results.admin_get_queryset", wraps=admin_get_queryset
    )
    data = get_product_flat(1, "annotated__is_null+1,size-2", {})
    assert data == [["NotNull", 2.0], ["NotNull", 1.0]]
    assert len(mock.call_args_list) == 2


def test_get_annotated_field_down_tree(products, get_product_flat, mocker):
    mock = mocker.patch(
        "data_browser.orm_results.admin_get_queryset", wraps=admin_get_queryset
    )
    data = get_product_flat(
        1,
        "producer__address__andrew+1,size-2",
        {"producer__address__andrew__not_equals": ["bad"]},
    )
    assert data == [["good", 1]]
    assert len(mock.call_args_list) == 2


@pytest.mark.skipif(django.VERSION < (2, 1), reason="Django version 2.1 required")
def test_get_aggregated_annotated_field_down_tree(products, get_product_flat, mocker):
    mock = mocker.patch(
        "data_browser.orm_results.admin_get_queryset", wraps=admin_get_queryset
    )
    data = get_product_flat(1, "producer__address__andrew__count+1,size-2", {})
    assert data == [[0, 2], [2, 1]]
    assert len(mock.call_args_list) == 2


def test_get_annotated_field_function_down_tree(products, get_product_flat, mocker):
    mock = mocker.patch(
        "data_browser.orm_results.admin_get_queryset", wraps=admin_get_queryset
    )
    data = get_product_flat(1, "producer__address__andrew__is_null+1,size-2", {})
    assert data == [["NotNull", 1.0], ["IsNull", 2.0]]
    assert len(mock.call_args_list) == 2


@pytest.mark.usefixtures("products")
def test_get_multiple_calculated_fields_on_admins(get_product_flat):
    data = get_product_flat(3, "producer__address__bob,producer__frank", {})
    sortedAssert(data, [[None, "frank"], ["err", "frank"], ["bob", "frank"]])


@pytest.mark.usefixtures("products")
def test_get_calculated_field(get_product_flat):
    data = get_product_flat(2, "producer__address__fred", {})
    sortedAssert(data, [[None], ["bad"], ["fred"]])


@pytest.mark.usefixtures("products")
def test_get_property(get_product_flat):
    data = get_product_flat(2, "producer__address__tom", {})
    sortedAssert(data, [[None], ["bad"], ["tom"]])


@pytest.mark.usefixtures("products")
def test_get_float_average(get_product_flat):
    data = get_product_flat(1, "boat__average", {})
    sortedAssert(data, [[2.3]])


@pytest.mark.usefixtures("products")
def test_get_float_sum(get_product_flat):
    data = get_product_flat(1, "boat__sum", {})
    sortedAssert(data, [[4.6]])


@pytest.mark.usefixtures("products")
def test_get_time_function(get_product_flat):
    data = get_product_flat(1, "created_time__year", {})
    assert data == [[timezone.now().year]]


@pytest.mark.usefixtures("products")
def test_filter_year_bounds(get_product_flat):
    data = get_product_flat(1, "size+1", {"created_time__year__equals": ["1"]})
    assert data == [[1], [2]]  # this errors on the filter and returns all values

    data = get_product_flat(1, "size+1", {"created_time__year__equals": ["2"]})
    assert data == []

    data = get_product_flat(1, "size+1", {"created_time__year__equals": ["9998"]})
    assert data == []

    data = get_product_flat(1, "size+1", {"created_time__year__equals": ["9999"]})
    assert data == [[1], [2]]  # this errors on the filter and returns all values


@pytest.mark.usefixtures("products")
def test_get_is_null_function(get_product_flat):
    data = get_product_flat(1, "name+0,producer__address__city__is_null", {})
    assert data == [["a", "NotNull"], ["b", "NotNull"], ["c", "IsNull"]]


def test_get_aggregated_is_null_function(get_product_flat):
    a = models.Producer.objects.create()
    b = models.Producer.objects.create()
    models.Product.objects.create(producer=a)
    models.Product.objects.create(producer=b)
    models.Product.objects.create(producer=b)
    data = get_product_flat(1, "producer__id__is_null,id__count", {})
    assert data == [["NotNull", 3]]


@pytest.mark.usefixtures("products")
def test_get_aggregate(get_product_flat):
    data = get_product_flat(1, "size_unit,id__count", {})
    assert data == [["g", 3]]


@pytest.mark.usefixtures("products")
def test_get_time_aggregate(get_product_flat):
    data = get_product_flat(1, "size_unit,created_time__count", {})
    assert data == [["g", 1]]


@pytest.mark.usefixtures("products")
def test_filter_and_get_aggregate(get_product_flat):
    data = get_product_flat(1, "size_unit,id__count", {"id__count__gt": [0]})
    assert data == [["g", 3]]


@pytest.mark.usefixtures("products")
def test_filter_aggregate(get_product_flat):
    data = get_product_flat(1, "size_unit", {"id__count__gt": [0]})
    assert data == [["g"]]


@pytest.mark.usefixtures("products")
def test_filter_sum_boolean(get_product_flat):
    data = get_product_flat(1, "size_unit", {"onsale__sum__gt": [100]})
    assert data == []


@pytest.mark.usefixtures("products")
def test_filter_average_boolean(get_product_flat):
    data = get_product_flat(1, "size_unit", {"onsale__average__gt": [100]})
    assert data == []


@pytest.mark.usefixtures("products")
def test_filter_aggregate_no_fields_filter(get_product_flat):
    # no group by -> no having
    data = get_product_flat(1, "onsale__sum", {"onsale__sum__lt": [0]})
    assert data == [[None]]


@pytest.mark.usefixtures("products")
def test_filter_aggregate_no_fields_filter_different(get_product_flat):
    # no group by -> no having
    data = get_product_flat(1, "onsale__sum", {"onsale__average__lt": [0]})
    assert data == [[None]]


@pytest.mark.usefixtures("products")
def test_get_aggregate_and_calculated_field(get_product_flat):
    data = get_product_flat(2, "is_onsale,id__count", {})
    sortedAssert(data, [["False", 1], ["False", 1], ["False", 1]])


@pytest.mark.usefixtures("products")
def test_get_time_aggregate_and_calculated_field(get_product_flat):
    data = get_product_flat(2, "is_onsale,created_time__count", {})
    sortedAssert(data, [["False", 1], ["False", 1], ["False", 1]])


@pytest.mark.usefixtures("products")
def test_get_only_aggregate(get_product_flat):
    data = get_product_flat(1, "id__count", {})
    assert data == [[3]]


@pytest.mark.usefixtures("products")
def test_get_results_empty(get_product_flat):
    data = get_product_flat(0, "", {})
    assert data == []


@pytest.mark.usefixtures("products")
def test_get_results_sort(get_product_flat):
    data = get_product_flat(1, "size+0,name-1,size_unit", {})
    assert data == [[1, "b", "g"], [1, "a", "g"], [2, "c", "g"]]


@pytest.mark.usefixtures("products")
def test_get_results_pks(get_product_flat):
    data = get_product_flat(1, "id", {})
    assert {d[0] for d in data} == set(
        models.Product.objects.values_list("id", flat=True)
    )


@pytest.mark.usefixtures("products")
def test_get_results_calculated_field(get_product_flat):
    # query + prefetch producer
    data = get_product_flat(2, "name+0,producer__name,is_onsale", {})
    assert data == [["a", "Bob", "False"], ["b", "Bob", "False"], ["c", "Bob", "False"]]


@pytest.mark.usefixtures("products")
def test_get_results_filtered(get_product_flat):
    data = get_product_flat(1, "size,name", {"name__equals": ["a"]})
    assert data == [[1, "a"]]


@pytest.mark.usefixtures("products")
def test_get_results_excluded(get_product_flat):
    data = get_product_flat(1, "size-0,name", {"name__not_equals": ["a"]})
    assert data == [[2, "c"], [1, "b"]]


@pytest.mark.usefixtures("products")
def test_get_results_multi_excluded(get_product_flat):
    data = get_product_flat(1, "size-0,name", {"name__not_equals": ["a", "c"]})
    assert data == [[1, "b"]]


@pytest.mark.usefixtures("products")
def test_get_results_collapsed(get_product_flat):
    data = get_product_flat(1, "size-0,size_unit", {})
    assert data == [[2, "g"], [1, "g"]]


@pytest.mark.usefixtures("products")
def test_get_results_null_filter(get_product_flat):
    data = get_product_flat(1, "name", {"onsale__is_null": ["IsNull"]})
    sortedAssert(data, [["a"], ["b"], ["c"]])
    data = get_product_flat(1, "name", {"onsale__is_null": ["NotNull"]})
    sortedAssert(data, [])


@pytest.mark.usefixtures("products")
def test_get_results_boolean_filter(get_product_flat):
    models.Product.objects.update(onsale=True)
    data = get_product_flat(1, "name", {"onsale__equals": ["True"]})
    sortedAssert(data, [["a"], ["b"], ["c"]])
    data = get_product_flat(1, "name", {"onsale__equals": ["true"]})
    sortedAssert(data, [["a"], ["b"], ["c"]])
    data = get_product_flat(1, "name", {"onsale__equals": ["False"]})
    sortedAssert(data, [])
    data = get_product_flat(1, "name", {"onsale__equals": ["false"]})
    sortedAssert(data, [])


@pytest.mark.usefixtures("products")
def test_get_results_string_filter(get_product_flat):
    data = get_product_flat(1, "name", {"producer__name__equals": ["Bob"]})
    sortedAssert(data, [["a"], ["b"], ["c"]])
    data = get_product_flat(1, "name", {"producer__name__equals": ["bob"]})
    sortedAssert(data, [["a"], ["b"], ["c"]])
    data = get_product_flat(1, "name", {"producer__name__equals": ["fred"]})
    sortedAssert(data, [])
    data = get_product_flat(1, "name", {"producer__name__regex": ["\\"]})
    sortedAssert(data, [["a"], ["b"], ["c"]])


@pytest.mark.usefixtures("products")
def test_get_results_basic_flat(get_product_flat):
    # just a query
    data = get_product_flat(1, "name+0,producer__address__city", {})
    assert data == [["a", "london"], ["b", "london"], ["c", None]]


@pytest.mark.usefixtures("products")
def test_get_results_one_to_one(get_product_flat):
    # just a query
    data = get_product_flat(1, "name+0,producer__address__producer__address__city", {})
    assert data == [["a", "london"], ["b", "london"], ["c", None]]


@pytest.mark.usefixtures("products")
def test_get_results_calculated_causes_query(get_product_flat):
    # query products, fetch objects for calculated
    data = get_product_flat(2, "name+0,is_onsale,producer__address__city", {})
    assert data == [
        ["a", "False", "london"],
        ["b", "False", "london"],
        ["c", "False", None],
    ]


def test_get_results_admin_causes_query(get_product_flat, products):
    # query products, fetch objects for calculated
    data = get_product_flat(2, "name+0,admin,producer__address__city", {})
    assert data == [
        ["a", admin_link(products[0]), "london"],
        ["b", admin_link(products[1]), "london"],
        ["c", admin_link(products[2]), None],
    ]


@pytest.mark.usefixtures("pivot_products")
def test_get_pivot(get_product_pivot):
    data = get_product_pivot(
        3, "created_time__year+0,&created_time__month+1,id__count", {}
    )
    assert data == {
        "body": [[[1], [3]], [[2], [4]]],
        "cols": [["January"], ["Feburary"]],
        "rows": [[2020], [2021]],
    }


@pytest.mark.usefixtures("pivot_products")
def test_get_pivot_multi_agg(get_product_pivot):
    data = get_product_pivot(
        3, "created_time__year+0,&created_time__month+1,size__count,size__max", {}
    )
    assert data == {
        "body": [[[1, 1], [3, 6]], [[2, 3], [4, 10]]],
        "cols": [["January"], ["Feburary"]],
        "rows": [[2020], [2021]],
    }


@pytest.mark.usefixtures("pivot_products")
def test_get_pivot_all(get_product_pivot):
    data = get_product_pivot(
        1, "&created_time__year+0, &created_time__month+1,id__count", {}
    )
    assert data == {
        "body": [[[1]], [[2]], [[3]], [[4]]],
        "cols": [
            [2020, "January"],
            [2020, "Feburary"],
            [2021, "January"],
            [2021, "Feburary"],
        ],
        "rows": [[]],
    }


def test_pivot_sorting_with_empty_cell(get_product_pivot):
    address = models.Address.objects.create(city="london", street="bad")
    producer = models.Producer.objects.create(name="Bob", address=address)
    datetimes = [
        # 2021, 1, 1 is notably missing
        datetime(2021, 2, 1, tzinfo=timezone.utc),
        datetime(2022, 1, 1, tzinfo=timezone.utc),
        datetime(2022, 1, 2, tzinfo=timezone.utc),
        datetime(2022, 2, 1, tzinfo=timezone.utc),
        datetime(2022, 2, 2, tzinfo=timezone.utc),
        datetime(2022, 2, 3, tzinfo=timezone.utc),
    ]
    for dt in datetimes:
        models.Product.objects.create(created_time=dt, name=str(dt), producer=producer)

    data = get_product_pivot(
        3, "&created_time__year+1,created_time__month+2,id__count", {}
    )
    assert data == {
        "body": [[[None], [1]], [[2], [3]]],
        "rows": [["January"], ["Feburary"]],
        "cols": [[2021], [2022]],
    }

    data = get_product_pivot(
        3, "&created_time__year+2,created_time__month+1,id__count", {}
    )
    assert data == {
        "body": [[[None], [1]], [[2], [3]]],
        "rows": [["January"], ["Feburary"]],
        "cols": [[2021], [2022]],
    }


def test_pivot_sorting_body(get_product_pivot):
    address = models.Address.objects.create(city="london", street="bad")
    producer = models.Producer.objects.create(name="Bob", address=address)
    datetimes = [
        datetime(2021, 1, 1, tzinfo=timezone.utc),
        datetime(2021, 2, 1, tzinfo=timezone.utc),
        datetime(2021, 2, 2, tzinfo=timezone.utc),
        datetime(2021, 2, 3, tzinfo=timezone.utc),
        datetime(2022, 1, 1, tzinfo=timezone.utc),
        datetime(2022, 1, 2, tzinfo=timezone.utc),
        datetime(2022, 1, 3, tzinfo=timezone.utc),
        datetime(2022, 1, 4, tzinfo=timezone.utc),
        datetime(2022, 2, 1, tzinfo=timezone.utc),
        datetime(2022, 2, 2, tzinfo=timezone.utc),
        datetime(2022, 2, 3, tzinfo=timezone.utc),
    ]
    for dt in datetimes:
        models.Product.objects.create(created_time=dt, name=str(dt), producer=producer)

    data = get_product_pivot(
        3, "&created_time__year+1,created_time__month+3,id__count+2", {}
    )
    assert data == {
        "body": [[[1], [3]], [[4], [3]]],
        "rows": [["January"], ["Feburary"]],
        "cols": [[2021], [2022]],
    }

    data = get_product_pivot(
        3, "&created_time__year+1,created_time__month+3,id__count-2", {}
    )
    assert data == {
        "body": [[[3], [1]], [[3], [4]]],
        "rows": [["Feburary"], ["January"]],
        "cols": [[2021], [2022]],
    }


@pytest.mark.usefixtures("pivot_products")
def test_pivot_having(get_product_pivot):
    data = get_product_pivot(
        3,
        "&created_time__year,created_time__month,id__count",
        {"id__count__equals": [4]},
    )
    assert data == {"body": [[[4]]], "rows": [["Feburary"]], "cols": [[2021]]}


jan = "January"
feb = "Feburary"
testdata = [
    # drcb, rows, cols, body
    # drcb = has_data, row_field, col_field, body_fields
    ("----", [], [], []),
    ("---b", [[0, None]], [[]], [[[]]]),
    ("--c-", [], [], []),
    ("--cb", [], [], []),
    ("-r--", [], [], []),
    ("-r-b", [], [], []),
    ("-rc-", [], [], []),
    ("-rcb", [], [], []),
    ("d---", [], [], []),
    ("d--b", [[10, 10]], [[]], [[[]]]),
    ("d-c-", [[]], [[jan], [feb]], [[[]], [[]]]),
    ("d-cb", [[]], [[jan], [feb]], [[[4, 6]], [[6, 10]]]),
    ("dr--", [[2020], [2021]], [[]], [[[], []]]),
    ("dr-b", [[2020, 3, 3], [2021, 7, 10]], [[]], [[[], []]]),
    ("drc-", [[2020], [2021]], [[jan], [feb]], [[[], []], [[], []]]),
    ("drcb", [[2020], [2021]], [[jan], [feb]], [[[1, 1], [3, 6]], [[2, 3], [4, 10]]]),
]


@pytest.mark.usefixtures("pivot_products")
@pytest.mark.parametrize("key,rows,cols,body", testdata)
def test_get_pivot_permutations(get_product_pivot, key, rows, cols, body):
    fields = []
    filters = {}
    if "r" in key:  # unpivoted field
        fields.append("created_time__year+0")
    if "c" in key:  # pivoted field
        fields.append("&created_time__month+1")
    if "b" in key:  # fields in the body
        fields.extend(["size__count", "size__max"])
    if "d" not in key:  # actually has data
        filters["id__equals"] = ["-1"]

    queries = 0 if key.endswith("---") else 1
    if "r" in key and "c" in key:
        queries += 2

    results = get_product_pivot(queries, ",".join(fields), filters)
    assert results["rows"] == rows
    assert results["cols"] == cols
    assert results["body"] == body


def test_get_fields(orm_models):

    # remap pk to id
    assert "pk" not in orm_models["core.Product"].fields
    assert "id" in orm_models["core.Product"].fields

    # no many to many fields
    assert "tags" not in orm_models["core.Product"].fields

    # no admin on inlines
    assert "admin" not in orm_models["core.InlineAdmin"].fields


class TestPermissions:
    def get_fields_with_perms(self, rf, perms):
        user = User.objects.create()
        for perm in perms:
            user.user_permissions.add(Permission.objects.get(codename=f"change_{perm}"))

        request = rf.get("/")
        request.user = user
        return get_models(request)

    def test_all_perms(self, rf, admin_user):
        orm_models = self.get_fields_with_perms(
            rf, ["normal", "notinadmin", "inadmin", "inlineadmin"]
        )

        assert "core.NotInAdmin" not in orm_models
        assert orm_models["core.InAdmin"] == OrmModel(
            fields=KEYS("admin", "id", "name"), admin=ANY(BaseModelAdmin)
        )
        assert orm_models["core.InlineAdmin"] == OrmModel(
            fields=KEYS("id", "name", "in_admin"), admin=ANY(BaseModelAdmin)
        )
        assert orm_models["core.Normal"] == OrmModel(
            fields=KEYS("admin", "id", "name", "in_admin", "inline_admin"),
            admin=ANY(BaseModelAdmin),
        )

    @pytest.mark.django_db
    def test_no_perms(self, rf):
        orm_models = self.get_fields_with_perms(rf, ["normal"])

        assert "core.NotInAdmin" not in orm_models
        assert "core.InAdmin" not in orm_models
        assert "core.InlineAdmin" not in orm_models
        assert orm_models["core.Normal"] == OrmModel(
            fields=KEYS("admin", "id", "name"), admin=ANY(BaseModelAdmin)
        )

    @pytest.mark.django_db
    def test_inline_perms(self, rf):
        orm_models = self.get_fields_with_perms(rf, ["normal", "inlineadmin"])

        assert "core.NotInAdmin" not in orm_models
        assert "core.InAdmin" not in orm_models
        assert "core.InlineAdmin" not in orm_models
        assert orm_models["core.Normal"] == OrmModel(
            fields=KEYS("admin", "id", "name"), admin=ANY(BaseModelAdmin)
        )

    @pytest.mark.django_db
    def test_admin_perms(self, rf):
        orm_models = self.get_fields_with_perms(rf, ["normal", "inadmin"])

        assert "core.NotInAdmin" not in orm_models
        assert orm_models["core.InAdmin"] == OrmModel(
            fields=KEYS("admin", "id", "name"), admin=ANY(BaseModelAdmin)
        )
        assert "core.InlineAdmin" not in orm_models
        assert orm_models["core.Normal"] == OrmModel(
            fields=KEYS("admin", "id", "name", "in_admin"), admin=ANY(BaseModelAdmin)
        )


@pytest.mark.django_db
@pytest.mark.parametrize(
    "lookup,value",
    [
        ("", "2020-01-02 03:04:05.000006"),
        ("year", 2020),
        ("quarter", 1),
        ("month", "January"),
        ("month_start", "2020-01-01"),
        ("day", 2),
        ("week_day", "Thursday"),
        ("hour", 3),
        ("minute", 4),
        ("second", 5),
        ("date", "2020-01-02"),
    ],
)
def test_all_datetime_functions(get_product_flat, lookup, value):
    models.Product.objects.create(
        producer=models.Producer.objects.create(),
        created_time=datetime(2020, 1, 2, 3, 4, 5, 6, tzinfo=timezone.utc),
    )

    fields = f"created_time__{lookup}" if lookup else "created_time"
    data = get_product_flat(1, fields, {})
    assert data == [[value]]


@pytest.mark.skipif(django.VERSION < (2, 2), reason="Django version 2.2 required")
@pytest.mark.django_db
@pytest.mark.parametrize(
    "lookup,value", [("iso_year", 2020), ("iso_week", 1), ("week_start", "2019-12-30")]
)
def test_all_datetime_functions_2_2(get_product_flat, lookup, value):
    models.Product.objects.create(
        producer=models.Producer.objects.create(),
        created_time=datetime(2020, 1, 2, 3, 4, 5, 6, tzinfo=timezone.utc),
    )

    fields = f"created_time__{lookup}" if lookup else "created_time"
    data = get_product_flat(1, fields, {})
    assert data == [[value]]


@pytest.mark.django_db
@pytest.mark.parametrize(
    "aggregation,value",
    [("count", 3), ("min", "2020-01-01 00:00:00"), ("max", "2020-01-03 00:00:00")],
)
def test_datetime_aggregations(get_product_flat, aggregation, value):
    producer = models.Producer.objects.create()
    models.Product.objects.create(
        producer=producer, created_time=datetime(2020, 1, 1, tzinfo=timezone.utc)
    )
    models.Product.objects.create(
        producer=producer, created_time=datetime(2020, 1, 2, tzinfo=timezone.utc)
    )
    models.Product.objects.create(
        producer=producer, created_time=datetime(2020, 1, 2, tzinfo=timezone.utc)
    )
    models.Product.objects.create(
        producer=producer, created_time=datetime(2020, 1, 3, tzinfo=timezone.utc)
    )

    data = get_product_flat(1, f"created_time__{aggregation}", {})
    assert data == [[value]]


@pytest.mark.django_db
@pytest.mark.parametrize(
    "aggregation,value", [("count", 3), ("min", "2020-01-01"), ("max", "2020-01-03")]
)
def test_date_aggregations(get_product_flat, aggregation, value):
    producer = models.Producer.objects.create()
    models.Product.objects.create(producer=producer, date=date(2020, 1, 1))
    models.Product.objects.create(producer=producer, date=date(2020, 1, 2))
    models.Product.objects.create(producer=producer, date=date(2020, 1, 2))
    models.Product.objects.create(producer=producer, date=date(2020, 1, 3))

    data = get_product_flat(1, f"date__{aggregation}", {})
    assert data == [[value]]


@pytest.mark.django_db
@pytest.mark.parametrize(
    "aggregation,value",
    [
        ("sum", "8:00:00"),
        ("average", "2:00:00"),
        ("count", 3),
        ("min", "1:00:00"),
        ("max", "3:00:00"),
    ],
)
def test_duration_aggregations(get_product_flat, aggregation, value):
    producer = models.Producer.objects.create()
    models.Product.objects.create(producer=producer, duration=timedelta(hours=1))
    models.Product.objects.create(producer=producer, duration=timedelta(hours=2))
    models.Product.objects.create(producer=producer, duration=timedelta(hours=2))
    models.Product.objects.create(producer=producer, duration=timedelta(hours=3))

    data = get_product_flat(1, f"duration__{aggregation}", {})
    assert data == [[value]]


def test_string_choice_field(get_product_flat):
    producer = models.Producer.objects.create()
    models.Product.objects.create(producer=producer, string_choice="a")
    models.Product.objects.create(producer=producer, string_choice="b")

    data = get_product_flat(1, f"string_choice", {"string_choice__equals": ["A"]})
    assert data == [["A"]]

    data = get_product_flat(
        1, f"string_choice__raw", {"string_choice__raw__equals": ["a"]}
    )
    assert data == [["a"]]


def test_number_choice_field(get_product_flat):
    producer = models.Producer.objects.create()
    models.Product.objects.create(producer=producer, number_choice=1)
    models.Product.objects.create(producer=producer, number_choice=2)

    data = get_product_flat(1, f"number_choice", {"number_choice__equals": ["A"]})
    assert data == [["A"]]

    data = get_product_flat(
        1, f"number_choice__raw", {"number_choice__raw__equals": [1]}
    )
    assert data == [[1]]


def test_is_null_variants(get_product_flat):
    producer = models.Producer.objects.create()
    models.Product.objects.create(producer=producer, size=1, onsale=True)
    models.Product.objects.create(producer=producer, size=2, onsale=None)

    assert get_product_flat(1, "onsale__is_null,size+1", {}) == [
        ["NotNull", 1],
        ["IsNull", 2],
    ]

    # __is_null=
    assert get_product_flat(1, "size", {"onsale__is_null": ["NotNull"]}) == [[1]]
    assert get_product_flat(1, "size", {"onsale__is_null": ["IsNull"]}) == [[2]]

    # __is_null__equals=
    assert get_product_flat(1, "size", {"onsale__is_null__equals": ["NotNull"]}) == [
        [1]
    ]
    assert get_product_flat(1, "size", {"onsale__is_null__equals": ["IsNull"]}) == [[2]]



# Generated by Django 3.1.1 on 2020-09-12 16:24

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models

import tests.core.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [("contenttypes", "0002_remove_content_type_name")]

    operations = [
        migrations.CreateModel(
            name="Address",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("city", models.TextField()),
                ("street", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="InAdmin",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="InlineAdmin",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                (
                    "in_admin",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="core.inadmin"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="NotInAdmin",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="Producer",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                (
                    "address",
                    models.OneToOneField(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core.address",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Product",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                ("size", models.IntegerField(default=0)),
                ("size_unit", models.TextField()),
                ("onsale", models.BooleanField(null=True)),
                ("image", models.FileField(upload_to="")),
                ("fake", tests.core.models.FakeField()),
                (
                    "created_time",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("only_in_list_view", models.TextField()),
                ("hidden_inline", models.TextField()),
                ("hidden_model", models.TextField()),
                ("extra_inline", models.TextField()),
                ("extra_model", models.TextField()),
                ("boat", models.FloatField(null=True)),
                ("duration", models.DurationField(null=True)),
                ("date", models.DateField(null=True)),
                ("not_in_admin", models.TextField()),
                (
                    "string_choice",
                    models.CharField(
                        choices=[("a", "A"), ("b", tests.core.models.Word("B"))],
                        max_length=8,
                    ),
                ),
                (
                    "number_choice",
                    models.IntegerField(
                        choices=[(1, "A"), (2, tests.core.models.Word("B"))], default=1
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Tag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="SKU",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                (
                    "product",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="core.product"
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="product",
            name="default_sku",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="core.sku",
                related_name="products",
            ),
        ),
        migrations.AddField(
            model_name="product",
            name="fk_not_in_admin",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="core.inadmin",
            ),
        ),
        migrations.AddField(
            model_name="product",
            name="model_not_in_admin",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="core.notinadmin",
            ),
        ),
        migrations.AddField(
            model_name="product",
            name="producer",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="core.producer"
            ),
        ),
        migrations.AddField(
            model_name="product",
            name="tags",
            field=models.ManyToManyField(to="core.Tag"),
        ),
        migrations.CreateModel(
            name="Normal",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                (
                    "in_admin",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="core.inadmin"
                    ),
                ),
                (
                    "inline_admin",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core.inlineadmin",
                    ),
                ),
                (
                    "not_in_admin",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core.notinadmin",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Ignored",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                (
                    "in_admin",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="core.inadmin"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="GenericInlineAdmin",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.TextField()),
                ("object_id", models.PositiveIntegerField()),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
        ),
    ]



import json

import pytest
from django.contrib.auth.models import Permission, User

from data_browser.common import MAKE_PUBLIC_CODENAME
from data_browser.models import View

from .util import ANY


@pytest.fixture
def view(admin_user):
    return View.objects.create(
        owner=admin_user,
        name="name",
        description="description",
        public=False,
        model_name="core.Product",
        fields="admin",
        query="name__contains=sql",
    )


@pytest.fixture
def other_user():
    return User.objects.create()


@pytest.fixture
def other_view(other_user):
    return View.objects.create(
        owner=other_user,
        name="name2",
        description="description2",
        public=False,
        model_name="core.Product",
        fields="admin",
        query="name__contains=sql",
    )


@pytest.fixture
def limited_user(admin_user):
    admin_user.is_superuser = False
    admin_user.save()

    admin_user.user_permissions.add(
        *Permission.objects.exclude(codename=MAKE_PUBLIC_CODENAME)
    )
    return admin_user


class TestViewList:
    def test_bad_methods(self, admin_client):
        resp = admin_client.patch("/data_browser/api/views/")
        assert resp.status_code == 400
        resp = admin_client.put("/data_browser/api/views/")
        assert resp.status_code == 400
        resp = admin_client.delete("/data_browser/api/views/")
        assert resp.status_code == 400

    def test_get(self, admin_client, view, other_view):
        resp = admin_client.get("/data_browser/api/views/")
        assert resp.status_code == 200
        assert resp.json() == [
            {
                "name": "name",
                "description": "description",
                "public": False,
                "model": "core.Product",
                "fields": "admin",
                "query": "name__contains=sql",
                "publicLink": "N/A",
                "googleSheetsFormula": "N/A",
                "link": "/query/core.Product/admin.html?name__contains=sql&limit=1000",
                "createdTime": ANY(str),
                "pk": view.pk,
                "limit": 1000,
            }
        ]

    def test_post(self, admin_client, admin_user):
        resp = admin_client.post(
            "/data_browser/api/views/",
            json.dumps(
                {
                    "name": "test",
                    "description": "lorem ipsum",
                    "public": True,
                    "model": "core.Product",
                    # leave the last two out just cause
                }
            ),
            content_type="application/json",
        )
        assert resp.status_code == 200

        view = View.objects.get()
        link = f"http://testserver/data_browser/view/{view.public_slug}.csv"

        assert resp.json() == {
            "name": "test",
            "description": "lorem ipsum",
            "public": True,
            "model": "core.Product",
            "fields": "",
            "query": "",
            "publicLink": link,
            "googleSheetsFormula": f'=importdata("{link}")',
            "link": "/query/core.Product/.html?&limit=1000",
            "createdTime": ANY(str),
            "pk": view.pk,
            "limit": 1000,
        }

        assert view.owner == admin_user
        assert view.name == "test"
        assert view.description == "lorem ipsum"
        assert view.public
        assert view.model_name == "core.Product"
        assert view.fields == ""
        assert view.query == ""

    def test_post_cant_make_public_without_perm(self, admin_client, limited_user):
        resp = admin_client.post(
            "/data_browser/api/views/",
            json.dumps(
                {
                    "name": "test",
                    "description": "lorem ipsum",
                    "public": True,
                    "model": "core.Product",
                    # leave the last two out just cause
                }
            ),
            content_type="application/json",
        )
        assert resp.status_code == 200
        assert not resp.json()["public"]
        view = View.objects.get()
        assert not view.public


class TestViewDetail:
    # can't see other users view

    def test_bad_methods(self, admin_client, view):
        resp = admin_client.post(f"/data_browser/api/views/{view.pk}/")
        assert resp.status_code == 400
        resp = admin_client.put(f"/data_browser/api/views/{view.pk}/")
        assert resp.status_code == 400

    def test_get(self, admin_client, view):
        resp = admin_client.get(f"/data_browser/api/views/{view.pk}/")
        assert resp.status_code == 200
        assert resp.json() == {
            "name": "name",
            "description": "description",
            "public": False,
            "model": "core.Product",
            "fields": "admin",
            "query": "name__contains=sql",
            "publicLink": "N/A",
            "googleSheetsFormula": "N/A",
            "link": "/query/core.Product/admin.html?name__contains=sql&limit=1000",
            "createdTime": ANY(str),
            "pk": view.pk,
            "limit": 1000,
        }

    def test_get_other_owner(self, admin_client, other_view):
        resp = admin_client.get(f"/data_browser/api/views/{other_view.pk}/")
        assert resp.status_code == 404

    def test_delete(self, admin_client, view):
        assert View.objects.exists()
        resp = admin_client.delete(f"/data_browser/api/views/{view.pk}/")
        assert resp.status_code == 204
        assert not View.objects.exists()

    def test_delete_other_owner(self, admin_client, other_view):
        assert View.objects.exists()
        resp = admin_client.delete(f"/data_browser/api/views/{other_view.pk}/")
        assert resp.status_code == 404
        other_view.refresh_from_db()

    def test_patch(self, admin_client, admin_user, view):
        resp = admin_client.patch(
            f"/data_browser/api/views/{view.pk}/",
            json.dumps(
                {
                    "name": "test",
                    "description": "lorem ipsum",
                    "public": True,
                    "model": "core.Product",
                    # leave the last two out just cause
                }
            ),
            content_type="application/json",
        )
        assert resp.status_code == 200

        view.refresh_from_db()
        link = f"http://testserver/data_browser/view/{view.public_slug}.csv"

        assert resp.json() == {
            "name": "test",
            "description": "lorem ipsum",
            "public": True,
            "model": "core.Product",
            "fields": "admin",
            "query": "name__contains=sql",
            "publicLink": link,
            "googleSheetsFormula": f'=importdata("{link}")',
            "link": "/query/core.Product/admin.html?name__contains=sql&limit=1000",
            "createdTime": ANY(str),
            "pk": view.pk,
            "limit": 1000,
        }

        assert view.owner == admin_user
        assert view.name == "test"
        assert view.description == "lorem ipsum"
        assert view.public
        assert view.model_name == "core.Product"
        assert view.fields == "admin"
        assert view.query == "name__contains=sql"

    def test_patch_other_owner(self, admin_client, other_user, other_view):
        resp = admin_client.patch(
            f"/data_browser/api/views/{other_view.pk}/",
            json.dumps({"name": "test", "description": "lorem ipsum"}),
            content_type="application/json",
        )
        assert resp.status_code == 404
        other_view.refresh_from_db()
        assert other_view.owner == other_user
        assert other_view.name == "name2"
        assert other_view.description == "description2"

    def test_patch_cant_make_public_without_perm(
        self, admin_client, limited_user, view
    ):
        resp = admin_client.patch(
            f"/data_browser/api/views/{view.pk}/",
            json.dumps({"public": True}),
            content_type="application/json",
        )
        assert resp.status_code == 200
        assert not resp.json()["public"]
        view.refresh_from_db()
        assert not view.public

    def test_patch_text_limit(self, admin_client, limited_user, view):
        resp = admin_client.patch(
            f"/data_browser/api/views/{view.pk}/",
            json.dumps({"limit": "bob"}),
            content_type="application/json",
        )
        assert resp.status_code == 200
        assert not resp.json()["public"]
        view.refresh_from_db()
        assert view.limit == 1

    def test_patch_negative_limit(self, admin_client, limited_user, view):
        resp = admin_client.patch(
            f"/data_browser/api/views/{view.pk}/",
            json.dumps({"limit": -1}),
            content_type="application/json",
        )
        assert resp.status_code == 200
        assert not resp.json()["public"]
        view.refresh_from_db()
        assert view.limit == 1



# Generated by Django 2.0.13 on 2020-05-31 13:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data_browser", "0005_auto_20200516_1726"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="view",
            options={
                "permissions": [
                    ("make_view_public", "Can make a saved view publicly available")
                ]
            },
        ),
        migrations.AlterField(
            model_name="view", name="query", field=models.TextField(blank=True),
        ),
    ]



import pytest

from data_browser import migration_helpers
from data_browser.models import View
from data_browser.orm_admin import get_models
from data_browser.query import BoundQuery


@pytest.fixture
def models(req):
    return get_models(req)


@pytest.mark.django_db
@pytest.mark.parametrize(
    "before,after",
    [
        # is_null
        ("name__is_null=True", "name__is_null=IsNull"),
        ("name__is_null=False", "name__is_null=NotNull"),
        ("name__is_null=true", "name__is_null=IsNull"),
        ("name__is_null=false", "name__is_null=NotNull"),
        # is_null equals
        ("name__is_null__equals=True", "name__is_null__equals=IsNull"),
        ("name__is_null__equals=False", "name__is_null__equals=NotNull"),
        ("name__is_null__equals=true", "name__is_null__equals=IsNull"),
        ("name__is_null__equals=false", "name__is_null__equals=NotNull"),
        # string_choice equals / not_equals
        ("string_choice__equals=a", "string_choice__equals=A"),
        ("string_choice__equals=c", "string_choice__raw__equals=c"),
        ("string_choice__not_equals=a", "string_choice__not_equals=A"),
        ("string_choice__not_equals=c", "string_choice__raw__not_equals=c"),
        # string_choice other
        ("string_choice__contains=a", "string_choice__raw__contains=a"),
        ("string_choice__starts_with=a", "string_choice__raw__starts_with=a"),
        ("string_choice__ends_with=a", "string_choice__raw__ends_with=a"),
        ("string_choice__regex=a", "string_choice__raw__regex=a"),
        ("string_choice__not_contains=a", "string_choice__raw__not_contains=a"),
        ("string_choice__not_starts_with=a", "string_choice__raw__not_starts_with=a"),
        ("string_choice__not_ends_with=a", "string_choice__raw__not_ends_with=a"),
        ("string_choice__not_regex=a", "string_choice__raw__not_regex=a"),
        # number_choice equals / not_equals
        ("number_choice__equals=1", "number_choice__equals=A"),
        ("number_choice__equals=3", "number_choice__raw__equals=3.0"),
        ("number_choice__not_equals=1", "number_choice__not_equals=A"),
        ("number_choice__not_equals=3", "number_choice__raw__not_equals=3.0"),
        # number_choice other
        ("number_choice__gt=1", "number_choice__raw__gt=1"),
        ("number_choice__gte=1", "number_choice__raw__gte=1"),
        ("number_choice__lt=1", "number_choice__raw__lt=1"),
        ("number_choice__lte=1", "number_choice__raw__lte=1"),
        # other good stuff
        ("name__equals=bob", "name__equals=bob"),
        ("name__contains=bob", "name__contains=bob"),
        ("boat__equals=1", "boat__equals=1"),
        ("boat__gt=1", "boat__gt=1"),
        # other bad stuff
        ("wtf__is_null=True", "wtf__is_null=True"),
        ("wtf__is_null__equals=True", "wtf__is_null__equals=True"),
        ("wtf__equals=a", "wtf__equals=a"),
        ("wtf__contains=bob", "wtf__contains=bob"),
        ("wtf__gt=1", "wtf__gt=1"),
        ("string_choice__wtf=a", "string_choice__raw__wtf=a"),
        ("number_choice__wtf=1", "number_choice__raw__wtf=1"),
        ("name__is_null=wtf", "name__is_null=wtf"),
        ("name__is_null__equals=wtf", "name__is_null__equals=wtf"),
    ],
)
def test_0009(models, before, after):
    valid = int("wtf" not in before)

    view = View.objects.create(model_name="core.Product", query=before)
    migration_helpers.forwards_0009(View)
    view.refresh_from_db()
    assert view.query == after
    assert len(BoundQuery.bind(view.get_query(), models).valid_filters) == valid

    view = View.objects.create(model_name="core.SKU", query=f"product__{before}")
    migration_helpers.forwards_0009(View)
    view.refresh_from_db()
    assert view.query == f"product__{after}"
    assert len(BoundQuery.bind(view.get_query(), models).valid_filters) == valid


def test_0009_multiple_filters(models):
    view = View.objects.create(
        model_name="core.Product",
        query="name__is_null=True&boat__gt=1&name__is_null__equals=True",
    )
    migration_helpers.forwards_0009(View)
    view.refresh_from_db()
    assert view.query == "name__is_null=IsNull&boat__gt=1&name__is_null__equals=IsNull"
    assert len(BoundQuery.bind(view.get_query(), models).valid_filters) == 3


@pytest.mark.django_db
def test_0009_no_filters():
    view = View.objects.create(model_name="core.Product")
    migration_helpers.forwards_0009(View)
    view.refresh_from_db()
    assert view.query == ""



from datetime import date, datetime, timedelta

import pytest
from django.http import QueryDict
from django.utils import timezone

from data_browser import orm_fields
from data_browser.orm_admin import OrmModel, get_fields_for_type
from data_browser.query import BoundQuery, Query, QueryField, QueryFilter
from data_browser.types import (
    ASC,
    DSC,
    BooleanType,
    DateTimeType,
    DateType,
    DurationType,
    HTMLType,
    IsNullType,
    NumberChoiceType,
    NumberType,
    StringArrayType,
    StringChoiceType,
    StringType,
    UnknownType,
)

from .util import ANY


@pytest.fixture
def query():
    return Query(
        "app.model",
        [
            QueryField("fa", False, ASC, 1),
            QueryField("fd", False, DSC, 0),
            QueryField("fn"),
        ],
        [QueryFilter("bob", "equals", "fred")],
    )


@pytest.fixture
def orm_models():
    return {
        "string": OrmModel(get_fields_for_type(StringType)),
        "number": OrmModel(get_fields_for_type(NumberType)),
        "app.model": OrmModel(
            fields={
                "fa": orm_fields.OrmConcreteField(
                    model_name="app.model",
                    name="fa",
                    pretty_name="fa",
                    type_=StringType,
                    rel_name=StringType.name,
                    choices=None,
                ),
                "fd": orm_fields.OrmConcreteField(
                    model_name="app.model",
                    name="fd",
                    pretty_name="fd",
                    type_=StringType,
                    rel_name=StringType.name,
                    choices=None,
                ),
                "fn": orm_fields.OrmCalculatedField(
                    model_name="app.model", name="fn", pretty_name="fn", func=None
                ),
                "bob": orm_fields.OrmConcreteField(
                    model_name="app.model",
                    name="bob",
                    pretty_name="bob",
                    type_=StringType,
                    rel_name=StringType.name,
                    choices=None,
                ),
                "num": orm_fields.OrmConcreteField(
                    model_name="app.model",
                    name="num",
                    pretty_name="num",
                    type_=NumberType,
                    rel_name=NumberType.name,
                    choices=None,
                ),
                "tom": orm_fields.OrmFkField(
                    model_name="app.model",
                    name="tom",
                    pretty_name="tom",
                    rel_name="app.Tom",
                ),
            },
            admin=True,
        ),
        "app.Tom": OrmModel(
            fields={
                "jones": orm_fields.OrmConcreteField(
                    model_name="app.Tom",
                    name="jones",
                    pretty_name="jones",
                    type_=StringType,
                    rel_name=StringType.name,
                    choices=None,
                ),
                "michael": orm_fields.OrmFkField(
                    model_name="app.Tom",
                    name="michael",
                    pretty_name="michael",
                    rel_name="app.Michael",
                ),
            },
            admin=True,
        ),
        "app.Michael": OrmModel(
            fields={
                "bolton": orm_fields.OrmConcreteField(
                    model_name="app.Michael",
                    name="bolton",
                    pretty_name="bolton",
                    type_=StringType,
                    rel_name=StringType.name,
                    choices=None,
                )
            },
            admin=True,
        ),
    }


@pytest.fixture
def bound_query(query, orm_models):
    return BoundQuery.bind(query, orm_models)


class TestQuery:
    def test_from_request(self, query):
        q = Query.from_request(
            "app.model", "fa+1,fd-0,fn", QueryDict("bob__equals=fred")
        )
        assert q == query

    def test_from_request_duplicate_field(self, query):
        q = Query.from_request(
            "app.model", "fa+1,fd-0,fn,&fa-2", QueryDict("bob__equals=fred")
        )
        assert q == query

    def test_from_request_with_limit(self, query):
        q = Query.from_request(
            "app.model", "fa+1,fd-0,fn", QueryDict("limit=123&bob__equals=fred")
        )
        query.limit = 123
        assert q == query

    def test_from_request_with_bad_limit(self, query):
        q = Query.from_request(
            "app.model", "fa+1,fd-0,fn", QueryDict("limit=bob&bob__equals=fred")
        )
        query.limit = 1000
        assert q == query

    def test_from_request_with_low_limit(self, query):
        q = Query.from_request(
            "app.model", "fa+1,fd-0,fn", QueryDict("limit=0&bob__equals=fred")
        )
        query.limit = 1
        assert q == query

    def test_from_request_with_related_filter(self):
        q = Query.from_request("app.model", "", QueryDict("bob__jones__equals=fred"))
        assert q == Query(
            "app.model", [], [QueryFilter("bob__jones", "equals", "fred")]
        )

    def test_from_request_with_missing(self):
        q = Query.from_request("app.model", ",,", QueryDict(""))
        assert q == Query("app.model", [], [])

    def test_from_request_filter_no_value(self):
        q = Query.from_request("app.model", "", QueryDict("joe__equals="))
        assert q == Query("app.model", [], [QueryFilter("joe", "equals", "")])

    def test_from_request_filter_no_lookup(self):
        q = Query.from_request("app.model", "", QueryDict("joe=tom"))
        assert q == Query("app.model", [], [])

    def test_from_request_filter_bad_lookup(self):
        q = Query.from_request("app.model", "", QueryDict("joe__blah=123"))
        assert q == Query("app.model", [], [QueryFilter("joe", "blah", "123")])

    def test_from_request_filter_no_name(self):
        q = Query.from_request("app.model", "", QueryDict("=123"))
        assert q == Query("app.model", [], [])

    def test_from_request_field_no_name(self):
        q = Query.from_request("app.model", "+2", QueryDict(""))
        assert q == Query("app.model", [QueryField("", False, ASC, 2)], [])

    def test_from_request_field_no_priority(self):
        q = Query.from_request("app.model", "fn+", QueryDict(""))
        assert q == Query("app.model", [QueryField("fn")], [])

    def test_from_request_field_bad_priority(self):
        q = Query.from_request("app.model", "fn+x", QueryDict(""))
        assert q == Query("app.model", [QueryField("fn")], [])

    def test_from_request_field_pivoted(self):
        q = Query.from_request("app.model", "&fn", QueryDict(""))
        assert q == Query("app.model", [QueryField("fn", True)], [])

    def test_url(self, query):
        query.limit = 123
        assert (
            query.get_url("html")
            == "/data_browser/query/app.model/fa+1,fd-0,fn.html?bob__equals=fred&limit=123"
        )

    def test_url_no_filters(self, query):
        query.filters = []
        assert (
            query.get_url("html")
            == "/data_browser/query/app.model/fa+1,fd-0,fn.html?limit=1000"
        )


class TestBoundQuery:
    def test_fields(self, bound_query):
        assert [f.path for f in bound_query.fields] == [["fa"], ["fd"], ["fn"]]

    def test_sort_fields(self, bound_query):
        assert [(f.path, f.direction, f.priority) for f in bound_query.sort_fields] == [
            (["fd"], DSC, 0),
            (["fa"], ASC, 1),
        ]

    def test_bad_field(self, orm_models):
        query = Query("app.model", [QueryField("yata")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_bad_field_lookup(self, orm_models):
        query = Query("app.model", [QueryField("fa__count__bob")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_bad_fk(self, orm_models):
        query = Query("app.model", [QueryField("yata__yata")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_bad_fk_field(self, orm_models):
        query = Query("app.model", [QueryField("tom__yata")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_bad_fk_field_aggregate(self, orm_models):
        query = Query("app.model", [QueryField("tom__jones__yata")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_bad_long_fk(self, orm_models):
        query = Query("app.model", [QueryField("yata__yata__yata")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_aggregate(self, orm_models):
        query = Query("app.model", [QueryField("tom__jones__count")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == [["tom", "jones", "count"]]

    def test_piovt_aggregate(self, orm_models):
        query = Query("app.model", [QueryField("tom__jones__count", pivoted=True)], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.pivoted for f in bound_query.fields] == [False]

    def test_piovt(self, orm_models):
        query = Query("app.model", [QueryField("tom__jones", pivoted=True)], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.pivoted for f in bound_query.fields] == [True]

    def test_bad_filter(self, orm_models):
        query = Query("app.model", [], [QueryFilter("yata", "equals", "fred")])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.filters] == []

    def test_bad_filter_value(self, orm_models):
        query = Query(
            "app.model",
            [],
            [QueryFilter("num", "equals", "fred"), QueryFilter("num", "equals", 1)],
        )
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.value for f in bound_query.filters] == ["fred", 1]
        assert [f.value for f in bound_query.valid_filters] == [1]

    def test_fk(self, orm_models):
        query = Query("app.model", [QueryField("tom")], [])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.fields] == []

    def test_filter_calculated_field(self, orm_models):
        query = Query("app.model", [], [QueryFilter("fn", "equals", "fred")])
        bound_query = BoundQuery.bind(query, orm_models)
        assert [f.path for f in bound_query.filters] == []


class TestType:
    def test_repr(self):
        assert repr(StringType) == f"StringType"


class TestStringType:
    @pytest.mark.django_db
    def test_validate(self):
        assert StringType.parse("contains", "hello", None) == ("hello", None)
        assert StringType.parse("pontains", "hello", None) == (None, ANY(str))
        assert StringType.parse("regex", ".*", None) == (".*", None)
        assert StringType.parse("regex", "\\", None) == (None, ANY(str))

    def test_default_lookup(self):
        assert StringType.default_lookup == "equals"

    def test_format(self):
        assert StringType.get_formatter(None)("bob") == "bob"
        assert StringType.get_formatter(None)(None) is None


class TestNumberType:
    def test_validate(self):
        assert NumberType.parse("gt", "6.1", None) == (6.1, None)
        assert NumberType.parse("pontains", "6.1", None) == (None, ANY(str))
        assert NumberType.parse("gt", "hello", None) == (None, ANY(str))
        assert NumberType.parse("is_null", "IsNull", None) == (True, None)
        assert NumberType.parse("is_null", "hello", None) == (None, ANY(str))

    def test_default_lookup(self):
        assert NumberType.default_lookup == "equals"

    def test_format(self):
        assert NumberType.get_formatter(None)(6) == 6
        assert NumberType.get_formatter(None)(None) is None


class TestDateTimeType:
    def test_validate(self):
        assert DateTimeType.parse("gt", "2018-03-20T22:31:23", None) == (
            ANY(datetime),
            None,
        )
        assert DateTimeType.parse("gt", "hello", None) == (None, ANY(str))
        assert DateTimeType.parse("pontains", "2018-03-20T22:31:23", None) == (
            None,
            ANY(str),
        )
        assert DateTimeType.parse("is_null", "IsNull", None) == (True, None)
        assert DateTimeType.parse("is_null", "hello", None) == (None, ANY(str))
        assert DateTimeType.parse("gt", "now", None) == (ANY(datetime), None)

        assert DateTimeType.parse("gt", "11-11-2018", None) == (ANY(datetime), None)
        assert DateTimeType.parse("gt", "11-22-2018", None) == (ANY(datetime), None)
        assert DateTimeType.parse("gt", "21-12-2018", None) == (ANY(datetime), None)
        assert DateTimeType.parse("gt", "11-12-2018", None) == (None, ANY(str))
        assert DateTimeType.parse("gt", "21-22-2018", None) == (None, ANY(str))

        assert DateTimeType.parse("gt", "2018-11-11", None) == (ANY(datetime), None)
        assert DateTimeType.parse("gt", "2018-11-22", None) == (ANY(datetime), None)
        assert DateTimeType.parse("gt", "2018-21-12", None) == (None, ANY(str))
        assert DateTimeType.parse("gt", "2018-11-12", None) == (ANY(datetime), None)
        assert DateTimeType.parse("gt", "2018-21-22", None) == (None, ANY(str))

        assert DateTimeType.parse("gt", "20181111", None) == (ANY(datetime), None)
        assert DateTimeType.parse("gt", "20181122", None) == (ANY(datetime), None)
        assert DateTimeType.parse("gt", "20182112", None) == (None, ANY(str))
        assert DateTimeType.parse("gt", "20181112", None) == (ANY(datetime), None)
        assert DateTimeType.parse("gt", "20182122", None) == (None, ANY(str))

        assert DateTimeType.parse("gt", "181111", None) == (None, ANY(str))
        assert DateTimeType.parse("gt", "181122", None) == (None, ANY(str))
        assert DateTimeType.parse("gt", "182112", None) == (None, ANY(str))
        assert DateTimeType.parse("gt", "181112", None) == (None, ANY(str))
        assert DateTimeType.parse("gt", "182122", None) == (None, ANY(str))

        assert DateTimeType.parse("gt", "111118", None) == (None, ANY(str))
        assert DateTimeType.parse("gt", "112218", None) == (ANY(datetime), None)
        assert DateTimeType.parse("gt", "211218", None) == (None, ANY(str))
        assert DateTimeType.parse("gt", "111218", None) == (None, ANY(str))
        assert DateTimeType.parse("gt", "212218", None) == (None, ANY(str))

    def test_default_lookup(self):
        assert DateTimeType.default_lookup == "equals"

    @pytest.mark.django_db
    def test_aware_format(self, settings):
        settings.USE_TZ = True
        assert (
            DateTimeType.get_formatter(None)(
                datetime(2020, 5, 19, 8, 42, 16, tzinfo=timezone.utc)
            )
            == "2020-05-19 08:42:16"
        )
        assert DateTimeType.get_formatter(None)(None) is None

    @pytest.mark.django_db
    def test_naive_format(self, settings):
        settings.USE_TZ = False
        assert (
            DateTimeType.get_formatter(None)(datetime(2020, 5, 19, 8, 42, 16))
            == "2020-05-19 08:42:16"
        )
        assert DateTimeType.get_formatter(None)(None) is None


class TestDurationType:
    def test_validate(self):
        assert DurationType.parse("gt", "5 days", None) == (timedelta(days=5), None)
        assert DurationType.parse("gt", "5 5:5", None) == (
            timedelta(days=5, hours=5, minutes=5),
            None,
        )
        assert DurationType.parse("gt", "5 dayss", None) == (None, ANY(str))
        assert DurationType.parse("pontains", "5 days", None) == (None, ANY(str))
        assert DurationType.parse("is_null", "IsNull", None) == (True, None)
        assert DurationType.parse("is_null", "hello", None) == (None, ANY(str))

    def test_default_lookup(self):
        assert DurationType.default_lookup == "equals"

    def test_format(self):
        assert (
            DurationType.get_formatter(None)(timedelta(days=5, minutes=6))
            == "5 days, 0:06:00"
        )
        assert DurationType.get_formatter(None)(timedelta(minutes=6)) == "0:06:00"
        assert DurationType.get_formatter(None)(None) is None


class TestDateType:
    def test_validate(self):
        assert DateType.parse("gt", "2018-03-20T22:31:23", None) == (ANY(date), None)
        assert DateType.parse("gt", "hello", None) == (None, ANY(str))
        assert DateType.parse("pontains", "2018-03-20T22:31:23", None) == (
            None,
            ANY(str),
        )
        assert DateType.parse("is_null", "IsNull", None) == (True, None)
        assert DateType.parse("is_null", "hello", None) == (None, ANY(str))
        assert DateType.parse("gt", "today", None) == (ANY(date), None)

        assert DateType.parse("gt", "11-22-2018", None) == (ANY(date), None)
        assert DateType.parse("gt", "21-12-2018", None) == (ANY(date), None)
        assert DateType.parse("gt", "11-12-2018", None) == (None, ANY(str))
        assert DateType.parse("gt", "21-22-2018", None) == (None, ANY(str))

    def test_default_lookup(self):
        assert DateType.default_lookup == "equals"

    def test_format(self):
        assert DateType.get_formatter(None)(date(2020, 5, 19)) == "2020-05-19"
        assert DateType.get_formatter(None)(None) is None


class TestBooleanType:
    def test_validate(self):
        assert BooleanType.parse("equals", "True", None) == (True, None)
        assert BooleanType.parse("equals", "False", None) == (False, None)
        assert BooleanType.parse("equals", "hello", None) == (None, ANY(str))
        assert BooleanType.parse("pontains", "True", None) == (None, ANY(str))

    def test_format(self):
        assert BooleanType.get_formatter(None)(None) is None

    def test_default_lookup(self):
        assert BooleanType.default_lookup == "equals"


class TestStringChoiceType:
    choices = [("a", "A"), ("b", "B"), ("c", "C")]

    def test_format(self):
        assert StringChoiceType.get_formatter(self.choices)("b") == "B"
        assert StringChoiceType.get_formatter(self.choices)(None) is None

    def test_bad_value(self):
        assert StringChoiceType.get_formatter(self.choices)("x") == "x"

    def test_validate(self):
        assert NumberChoiceType.parse("equals", "B", self.choices) == ("b", None)
        assert NumberChoiceType.parse("equals", "X", self.choices) == (None, ANY(str))


class TestNumberChoiceType:
    choices = [(1, "A"), (2, "B"), (3, "C")]

    def test_format(self):
        assert NumberChoiceType.get_formatter(self.choices)(2) == "B"
        assert NumberChoiceType.get_formatter(self.choices)(None) is None

    def test_bad_value(self):
        assert NumberChoiceType.get_formatter(self.choices)(6) == 6

    def test_validate(self):
        assert NumberChoiceType.parse("equals", "B", self.choices) == (2, None)
        assert NumberChoiceType.parse("equals", "X", self.choices) == (None, ANY(str))


class TestHTMLType:
    def test_format(self):
        assert HTMLType.get_formatter(None)(None) is None


class TestIsNullType:
    def test_format(self):
        assert IsNullType.get_formatter(IsNullType.choices)(True) == "IsNull"
        assert IsNullType.get_formatter(IsNullType.choices)(False) == "NotNull"
        assert IsNullType.get_formatter(IsNullType.choices)(None) == "IsNull"

    def test_validate(self):
        assert IsNullType.parse("equals", "IsNull", IsNullType.choices) == (True, None)
        assert IsNullType.parse("equals", "NotNull", IsNullType.choices) == (
            False,
            None,
        )


class TestUnknownType:
    def test_format(self):
        assert UnknownType.get_formatter(None)(None) is None


class TestArrayStringType:
    def test_validate(self):
        assert StringArrayType.parse("equals", '["X"]', None) == (["X"], None)
        assert StringArrayType.parse("equals", '"X"', None) == (None, ANY(str))



from collections import defaultdict

from django.db import models
from django.db.models import DurationField, IntegerField
from django.db.models.functions import Cast

from .orm_fields import OrmBaseField, OrmBoundField
from .types import (
    BooleanType,
    DateTimeType,
    DateType,
    DurationType,
    NumberType,
    StringChoiceType,
    StringType,
)
from .util import s

_TYPE_AGGREGATES = defaultdict(
    lambda: [("count", NumberType)],
    {
        StringType: [("count", NumberType)],
        StringChoiceType: [("count", NumberType)],
        NumberType: [
            ("average", NumberType),
            ("count", NumberType),
            ("max", NumberType),
            ("min", NumberType),
            ("std_dev", NumberType),
            ("sum", NumberType),
            ("variance", NumberType),
        ],
        DateTimeType: [
            ("count", NumberType),
            ("max", DateTimeType),
            ("min", DateTimeType),
        ],
        DateType: [("count", NumberType), ("max", DateType), ("min", DateType)],
        DurationType: [
            ("count", NumberType),
            ("average", DurationType),
            ("sum", DurationType),
            ("max", DurationType),
            ("min", DurationType),
        ],
        BooleanType: [("average", NumberType), ("sum", NumberType)],
    },
)


class _CastDuration(Cast):
    def __init__(self, expression):
        super().__init__(expression, output_field=DurationField())

    def as_mysql(self, compiler, connection, **extra_context):  # pragma: mysql
        # https://github.com/django/django/pull/13398
        template = "%(function)s(%(expressions)s AS signed integer)"
        return self.as_sql(compiler, connection, template=template, **extra_context)


def _get_django_aggregate(field_type, name):
    if field_type == BooleanType:
        return {
            "average": lambda x: models.Avg(Cast(x, output_field=IntegerField())),
            "sum": lambda x: models.Sum(Cast(x, output_field=IntegerField())),
        }[name]
    if field_type == DurationType and name in ["average", "sum"]:
        return {
            "average": lambda x: models.Avg(_CastDuration(x)),
            "sum": lambda x: models.Sum(_CastDuration(x)),
        }[name]
    else:
        return {
            # these all have result type number
            "average": models.Avg,
            "count": lambda x: models.Count(x, distinct=True),
            "max": models.Max,
            "min": models.Min,
            "std_dev": models.StdDev,
            "sum": models.Sum,
            "variance": models.Variance,
        }[name]


class OrmAggregateField(OrmBaseField):
    def __init__(self, model_name, name, type_):
        super().__init__(
            model_name, name, name.replace("_", " "), type_=type_, concrete=True
        )

    def bind(self, previous):
        assert previous
        queryset_path = previous.queryset_path + [self.name]
        agg_func = _get_django_aggregate(previous.type_, self.name)
        return OrmBoundField(
            field=self,
            previous=previous,
            full_path=previous.full_path + [self.name],
            pretty_path=previous.pretty_path + [self.pretty_name],
            queryset_path=queryset_path,
            aggregate_clause=(s(queryset_path), agg_func(s(previous.queryset_path))),
            having=True,
        )


def get_aggregates_for_type(type_):
    return {
        aggregate: OrmAggregateField(type_.name, aggregate, res_type)
        for aggregate, res_type in _TYPE_AGGREGATES[type_]
    }



import cProfile
import csv
import io
import itertools
import json
import marshal
import pstats
import sys

import django.contrib.admin.views.decorators as admin_decorators
import sqlparse
from django import http
from django.core.serializers.json import DjangoJSONEncoder
from django.shortcuts import get_object_or_404
from django.template import engines, loader
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils import timezone
from django.views.decorators import csrf

from . import version
from .common import HttpResponse, JsonResponse, can_make_public, settings
from .models import View
from .orm_admin import OPEN_IN_ADMIN, get_models
from .orm_results import get_result_queryset, get_results
from .query import BoundQuery, Query, QueryFilter
from .types import TYPES


def _get_query_data(bound_query):
    return {
        "filters": [
            {
                "pathStr": filter_.path_str,
                "lookup": filter_.lookup,
                "value": filter_.value,
            }
            for filter_ in bound_query.filters
        ],
        "filterErrors": [filter_.err_message for filter_ in bound_query.filters],
        "fields": [
            {
                "pathStr": field.path_str,
                "sort": field.direction,
                "priority": field.priority,
                "pivoted": field.pivoted,
            }
            for field in bound_query.fields
        ],
        "model": bound_query.model_name,
        "limit": bound_query.limit,
    }


def _get_model_fields(model_name, orm_models):
    orm_model = orm_models[model_name]

    def sort_model_fields(fields):
        front = {"id": 1, OPEN_IN_ADMIN: 2}
        return sorted(fields, key=lambda f: (front.get(f, sys.maxsize), f))

    all_fields = {
        name: {
            "model": orm_field.rel_name,
            "type": orm_field.type_.name if orm_field.type_ else None,
            "concrete": orm_field.concrete,
            "canPivot": orm_field.can_pivot,
            "prettyName": orm_field.pretty_name,
            "choices": [v for k, v in orm_field.choices],
            "defaultSort": orm_field.default_sort,
        }
        for name, orm_field in orm_model.fields.items()
    }

    q = Query(model_name, [], [QueryFilter(*f) for f in orm_model.default_filters])
    bq = BoundQuery.bind(q, orm_models)

    return {
        "fields": all_fields,
        "sortedFields": sort_model_fields(all_fields),
        "defaultFilters": [
            {
                "pathStr": filter_.path_str,
                "lookup": filter_.lookup,
                "value": filter_.value,
            }
            for filter_ in bq.filters
        ],
    }


def _get_config(request):
    orm_models = get_models(request)
    types = {
        name: {
            "lookups": {n: {"type": t} for n, t in type_.lookups.items()},
            "sortedLookups": list(type_.lookups),
            "defaultLookup": type_.default_lookup,
            "defaultValue": type_.default_value,
        }
        for name, type_ in TYPES.items()
    }

    all_model_fields = {
        model_name: _get_model_fields(model_name, orm_models)
        for model_name in orm_models
    }

    return {
        "baseUrl": reverse("data_browser:home"),
        "types": types,
        "allModelFields": all_model_fields,
        "sortedModels": sorted(
            name for name, model in orm_models.items() if model.root
        ),
        "canMakePublic": can_make_public(request.user),
        "sentryDsn": settings.DATA_BROWSER_FE_DSN,
        "defaultRowLimit": settings.DATA_BROWSER_DEFAULT_ROW_LIMIT,
    }


@admin_decorators.staff_member_required
def query_ctx(request, *, model_name="", fields=""):
    config = _get_config(request)
    return JsonResponse(config)


@csrf.ensure_csrf_cookie
@admin_decorators.staff_member_required
def query_html(request, *, model_name="", fields=""):
    config = _get_config(request)
    config = json.dumps(config, cls=DjangoJSONEncoder)
    config = (
        config.replace("<", "\\u003C").replace(">", "\\u003E").replace("&", "\\u0026")
    )

    if settings.DATA_BROWSER_DEV:  # pragma: no cover
        try:
            response = _get_from_js_dev_server(request)
        except Exception as e:
            return HttpResponse(f"Error loading from JS dev server.<br><br>{e}")

        template = engines["django"].from_string(response.text)
    else:
        template = loader.get_template("data_browser/index.html")

    return TemplateResponse(request, template, {"config": config, "version": version})


@admin_decorators.staff_member_required
def query(request, *, model_name, fields="", media):
    if media.startswith("profile") or media.startswith("pstats"):
        if "_" in media:
            prof_media, media = media.split("_")
        else:
            prof_media = media
            media = "json"

        profiler = cProfile.Profile()
        try:
            profiler.enable()

            query = Query.from_request(model_name, fields, request.GET)
            for x in _data_response(request, query, media, privileged=True):
                pass

            profiler.disable()

            if prof_media == "profile":
                buffer = io.StringIO()
                stats = pstats.Stats(profiler, stream=buffer)
                stats.sort_stats("cumulative").print_stats(50)
                stats.sort_stats("time").print_stats(50)
                buffer.seek(0)
                return HttpResponse(buffer, content_type="text/plain")
            elif prof_media == "pstats":
                buffer = io.BytesIO()
                marshal.dump(pstats.Stats(profiler).stats, buffer)
                buffer.seek(0)
                response = HttpResponse(buffer, content_type="application/octet-stream")
                response[
                    "Content-Disposition"
                ] = f"attachment; filename={query.model_name}-{timezone.now().isoformat()}.pstats"
                return response
            else:
                raise http.Http404(f"Bad file format {prof_media} requested")
        finally:
            if profiler:  # pragma: no branch
                profiler.disable()
    else:
        query = Query.from_request(model_name, fields, request.GET)
        return _data_response(request, query, media, privileged=True)


def view(request, pk, media):
    view = get_object_or_404(View.objects.filter(public=True), public_slug=pk)
    if (
        # some of these are checked by the admin but this is a good time to be paranoid
        view.owner.is_active
        and view.owner.is_staff
        and can_make_public(view.owner)
        and settings.DATA_BROWSER_ALLOW_PUBLIC
    ):
        request.user = view.owner  # public views are run as the person who owns them
        query = view.get_query()
        return _data_response(request, query, media, privileged=False)
    else:
        raise http.Http404("No View matches the given query.")


def pad(x):
    return [None] * max(0, x)


def concat(*lists):
    return itertools.chain.from_iterable(lists)


def flip_table(table):
    return (list(x) for x in zip(*table))


def join_tables(*tables):
    return (list(concat(*ts)) for ts in zip(*tables))


def format_table(fields, table, spacing=0):
    return concat(
        [[" ".join(f.pretty_path) for f in fields]],
        *[
            [[(row[f.path_str] if row else "") for f in fields]]
            + [pad(len(fields))] * spacing
            for row in table
        ],
    )


def pad_table(x, table):
    p = pad(x)
    return (p + row for row in table)


class Echo:
    def write(self, value):
        return value


def _data_response(request, query, media, privileged=False):
    orm_models = get_models(request)
    if query.model_name not in orm_models:
        raise http.Http404(f"{query.model_name} does not exist")
    bound_query = BoundQuery.bind(query, orm_models)

    if media == "csv":
        results = get_results(request, bound_query, orm_models, False)

        def csv_rows():
            # the pivoted column headers
            yield from pad_table(
                len(bound_query.row_fields) - 1,
                flip_table(
                    format_table(
                        bound_query.col_fields,
                        results["cols"],
                        spacing=len(bound_query.data_fields) - 1,
                    )
                ),
            )

            # the row headers and data area
            yield from pad_table(
                1 - len(bound_query.row_fields),
                join_tables(
                    format_table(bound_query.row_fields, results["rows"]),
                    *(
                        format_table(bound_query.data_fields, sub_table)
                        for sub_table in results["body"]
                    ),
                ),
            )

        writer = csv.writer(Echo())
        response = http.StreamingHttpResponse(
            (writer.writerow(row) for row in csv_rows()), content_type="text/csv"
        )
        response[
            "Content-Disposition"
        ] = f"attachment; filename={query.model_name}-{timezone.now().isoformat()}.csv"
        return response
    elif media == "json":
        results = get_results(request, bound_query, orm_models, True)
        resp = _get_query_data(bound_query) if privileged else {}
        resp.update(results)
        return JsonResponse(resp)
    elif privileged and media == "query":
        resp = _get_query_data(bound_query)
        return JsonResponse(resp)
    elif privileged and media in ["sql", "explain"]:
        query_set = get_result_queryset(request, bound_query, orm_models)
        if isinstance(query_set, list):
            res = "Not available for pure aggregates"
        else:
            if media == "sql":
                res = sqlparse.format(
                    str(query_set.query), reindent=True, keyword_case="upper"
                )
            elif media == "explain":
                res = query_set.explain()
            else:
                assert False
        return HttpResponse(res, content_type="text/plain")
    else:
        raise http.Http404(f"Bad file format {media} requested")


def _get_from_js_dev_server(request):  # pragma: no cover
    import requests

    upstream_url = f"http://127.0.0.1:3000{request.path}"
    print(f"Proxy request: {upstream_url}")
    method = request.META["REQUEST_METHOD"].lower()
    return getattr(requests, method)(upstream_url, stream=True)


@csrf.csrf_exempt
def proxy_js_dev_server(request, path):  # pragma: no cover
    """
    Proxy HTTP requests to the frontend dev server in development.

    The implementation is very basic e.g. it doesn't handle HTTP headers.

    """
    response = _get_from_js_dev_server(request)
    return http.StreamingHttpResponse(
        streaming_content=response.iter_content(2 ** 12),
        content_type=response.headers.get("Content-Type"),
        status=response.status_code,
        reason=response.reason,
    )



# Generated by Django 2.2.16 on 2020-10-26 20:25

from django.db import migrations

from data_browser.migration_helpers import forwards_0009


def forwards(apps, schema_editor):
    View = apps.get_model("data_browser", "View")
    forwards_0009(View)


class Migration(migrations.Migration):
    dependencies = [("data_browser", "0008_view_limit")]
    operations = [migrations.RunPython(forwards, migrations.RunPython.noop)]



from collections import defaultdict

import django
from django.db.models import (
    BooleanField,
    DateField,
    ExpressionWrapper,
    OuterRef,
    Q,
    Subquery,
    functions,
)

from .orm_fields import OrmBaseField, OrmBoundField
from .types import (
    ASC,
    DateTimeType,
    DateType,
    IsNullType,
    NumberChoiceType,
    NumberType,
    StringType,
)
from .util import s

_DATE_FUNCTIONS = [
    "is_null",
    "year",
    "quarter",
    "month",
    "day",
    "week_day",
    "month_start",
]
if django.VERSION >= (2, 2):  # pragma: no branch
    _DATE_FUNCTIONS += ["iso_year", "iso_week", "week_start"]


_TYPE_FUNCTIONS = defaultdict(
    lambda: ["is_null"],
    {
        DateType: _DATE_FUNCTIONS,
        DateTimeType: _DATE_FUNCTIONS + ["hour", "minute", "second", "date"],
        StringType: ["is_null", "length"],
    },
)


_month_choices = [
    (1, "January"),
    (2, "Feburary"),
    (3, "March"),
    (4, "April"),
    (5, "May"),
    (6, "June"),
    (7, "July"),
    (8, "August"),
    (9, "September"),
    (10, "October"),
    (11, "November"),
    (12, "December"),
]


_weekday_choices = [
    (1, "Sunday"),
    (2, "Monday"),
    (3, "Tuesday"),
    (4, "Wednesday"),
    (5, "Thursday"),
    (6, "Friday"),
    (7, "Saturday"),
]


def _get_django_function(name, qs):
    def IsNull(field_name):
        # https://code.djangoproject.com/ticket/32200
        if django.VERSION[:3] == (3, 1, 3):
            return Subquery(
                qs.annotate(
                    ddb_is_null=ExpressionWrapper(
                        Q(**{field_name: None}), output_field=BooleanField()
                    )
                )
                .filter(pk=OuterRef("id"))
                .values("ddb_is_null")[:1],
                output_field=BooleanField(),
            )
        else:  # pragma: django != 3.1.3
            return ExpressionWrapper(
                Q(**{field_name: None}), output_field=BooleanField()
            )

    mapping = {
        "year": (functions.ExtractYear, NumberType, (), ASC, {"useGrouping": False}),
        "quarter": (functions.ExtractQuarter, NumberType, (), ASC, {}),
        "month": (functions.ExtractMonth, NumberChoiceType, _month_choices, ASC, {}),
        "month_start": (
            lambda x: functions.TruncMonth(x, DateField()),
            DateType,
            (),
            ASC,
            {},
        ),
        "day": (functions.ExtractDay, NumberType, (), ASC, {}),
        "week_day": (
            functions.ExtractWeekDay,
            NumberChoiceType,
            _weekday_choices,
            ASC,
            {},
        ),
        "hour": (functions.ExtractHour, NumberType, (), ASC, {}),
        "minute": (functions.ExtractMinute, NumberType, (), ASC, {}),
        "second": (functions.ExtractSecond, NumberType, (), ASC, {}),
        "date": (functions.TruncDate, DateType, (), ASC, {}),
        "is_null": (IsNull, IsNullType, (), None, {}),
        "length": (functions.Length, NumberType, (), None, {}),
    }
    if django.VERSION >= (2, 2):  # pragma: no branch
        mapping.update(
            {
                "iso_year": (functions.ExtractIsoYear, NumberType, (), ASC, {}),
                "iso_week": (functions.ExtractWeek, NumberType, (), ASC, {}),
                "week_start": (
                    lambda x: functions.TruncWeek(x, DateField()),
                    DateType,
                    (),
                    ASC,
                    {},
                ),
            }
        )
    return mapping[name]


class OrmBoundFunctionField(OrmBoundField):
    def _annotate(self, request, qs):
        func = _get_django_function(self.name, qs)[0](s(self.previous.queryset_path))
        return qs.annotate(**{s(self.queryset_path): func})

    def parse(self, lookup, value):
        parsed, err_message = super().parse(lookup, value)
        if self.name in ["year", "iso_year"] and parsed is not None:
            if parsed < 2:
                err_message = "Can't filter to years less than 2"
            if parsed > 9998:
                err_message = "Can't filter to years greater than 9998"
        return parsed, err_message


class OrmFunctionField(OrmBaseField):
    def __init__(self, model_name, name, type_, choices, default_sort, format_hints):
        super().__init__(
            model_name,
            name,
            name.replace("_", " "),
            type_=type_,
            concrete=True,
            can_pivot=True,
            choices=choices,
            default_sort=default_sort,
            format_hints=format_hints,
        )

    def bind(self, previous):
        assert previous
        return OrmBoundFunctionField(
            field=self,
            previous=previous,
            full_path=previous.full_path + [self.name],
            pretty_path=previous.pretty_path + [self.pretty_name],
            queryset_path=previous.queryset_path + [self.name],
            filter_=True,
        )


def get_functions_for_type(type_):
    return {
        func: OrmFunctionField(type_.name, func, *_get_django_function(func, None)[1:])
        for func in _TYPE_FUNCTIONS[type_]
    }



# Generated by Django 2.0.13 on 2020-06-14 10:37

from django.db import migrations, models

import data_browser.models


class Migration(migrations.Migration):

    dependencies = [
        ("data_browser", "0006_auto_20200531_1450"),
    ]

    operations = [
        migrations.AddField(
            model_name="view",
            name="public_slug",
            field=models.CharField(default=data_browser.models.get_id, max_length=12),
        ),
    ]



import pytest
from django.contrib import admin
from django.contrib.admin.utils import flatten_fieldsets
from django.contrib.auth.models import Permission, User

from data_browser.admin import ViewAdmin
from data_browser.models import View


@pytest.fixture
def view(admin_user):
    return View(model_name="app.model", fields="fa+0,fd-1,fn", query="bob__equals=fred")


def test_open_view(view, rf):
    expected = '<a href="/data_browser/query/app.model/fa+0,fd-1,fn.html?bob__equals=fred&limit=1000">view</a>'
    assert ViewAdmin.open_view(view) == expected


def test_add_has_user(admin_client):
    user = User.objects.get()
    res = admin_client.get("/admin/data_browser/view/add/")
    assert res.status_code == 200
    assert res.context[0]["adminform"].form.initial["owner"] == user.pk


def test_change_form_links_have_full_url(view, admin_client):
    view.public = True
    view.save()
    res = admin_client.get(
        f"http://testserver/admin/data_browser/view/{view.pk}/change/"
    )
    assert res.status_code == 200

    all_fields = [f for fs in res.context[0]["adminform"] for l in fs for f in l]
    for field in all_fields:
        if field.is_readonly and field.field["name"] == "public_link":
            expected = f"http://testserver/data_browser/view/{view.public_slug}.csv"
            assert field.contents() == expected
            break
    else:
        assert False


@pytest.fixture
def get_admin_details(rf):
    def helper(admin_user, obj):
        request = rf.get("/")
        request.user = admin_user
        view_admin = ViewAdmin(View, admin.site)
        fields = set(flatten_fieldsets(view_admin.get_fieldsets(request, obj)))
        read_only = set(view_admin.get_readonly_fields(request, obj))
        assert fields == read_only
        return fields

    return helper


@pytest.fixture
def staff_user(admin_user):
    admin_user.is_superuser = False
    admin_user.user_permissions.add(Permission.objects.get(codename=f"change_view"))
    return admin_user


class TestAdminFieldsSuperUser:
    def test_add_page_edit_everything(self, admin_user, get_admin_details):
        fields = get_admin_details(admin_user, None)
        assert fields == {
            "description",
            "fields",
            "model_name",
            "name",
            "owner",
            "public",
            "public_slug",
            "query",
            "created_time",
            "google_sheets_formula",
            "id",
            "open_view",
            "public_link",
            "limit",
        }

    def test_private_view_edit_everything(self, admin_user, get_admin_details, view):
        fields = get_admin_details(admin_user, view)
        assert fields == {
            "description",
            "fields",
            "model_name",
            "name",
            "owner",
            "public",
            "public_slug",
            "query",
            "created_time",
            "google_sheets_formula",
            "id",
            "open_view",
            "public_link",
            "limit",
        }

    def test_public_view_edit_everything(self, admin_user, get_admin_details, view):
        view.public = True
        fields = get_admin_details(admin_user, view)
        assert fields == {
            "description",
            "fields",
            "model_name",
            "name",
            "owner",
            "public",
            "public_slug",
            "query",
            "created_time",
            "google_sheets_formula",
            "id",
            "open_view",
            "public_link",
            "limit",
        }


class TestAdminFieldsStaffUser:
    def test_add_page_no_public_fields(self, staff_user, get_admin_details):
        fields = get_admin_details(staff_user, None)
        assert fields == {
            "description",
            "fields",
            "model_name",
            "name",
            "owner",
            "query",
            "created_time",
            "id",
            "open_view",
            "limit",
        }

    def test_private_view_no_public_fields(self, staff_user, get_admin_details, view):
        fields = get_admin_details(staff_user, view)
        assert fields == {
            "description",
            "fields",
            "model_name",
            "name",
            "owner",
            "query",
            "created_time",
            "id",
            "open_view",
            "limit",
        }

    def test_public_view_readonly(self, staff_user, get_admin_details, view):
        view.public = True
        fields = get_admin_details(staff_user, view)
        assert fields == {
            "created_time",
            "description",
            "fields",
            "id",
            "model_name",
            "name",
            "open_view",
            "owner",
            "query",
            "limit",
        }



import urllib
from dataclasses import dataclass
from typing import Any, Optional, Sequence

from django.urls import reverse
from django.utils.functional import cached_property

from .common import settings
from .types import ASC, DSC


@dataclass
class QueryField:
    path: str
    pivoted: bool = False
    direction: str = None
    priority: int = None

    def __post_init__(self):
        self.path = self.path.split("__")
        assert (self.direction is None) == (self.priority is None)


@dataclass
class QueryFilter:
    path: str
    lookup: str
    value: str

    def __post_init__(self):
        self.path = self.path.split("__")


def parse_sort(value, symbol, direction):
    path, priority = value.split(symbol)
    try:
        return path, direction, int(priority)
    except:  # noqa: E722  input sanitization
        return path, None, None


@dataclass
class Query:
    model_name: str
    fields: Sequence[QueryField]
    filters: Sequence[QueryFilter]
    limit: int = settings.DATA_BROWSER_DEFAULT_ROW_LIMIT

    @classmethod
    def from_request(cls, model_name, field_str, get_args):
        fields = []
        found_fields = set()
        for field in field_str.split(","):
            field = field.strip()
            if field:
                # pivot
                if field.startswith("&"):
                    field = field[1:]
                    pivoted = True
                else:
                    pivoted = False

                # sort
                if "+" in field:
                    path, direction, priority = parse_sort(field, "+", ASC)
                elif "-" in field:
                    path, direction, priority = parse_sort(field, "-", DSC)
                else:
                    path, direction, priority = field, None, None

                # collect
                if path not in found_fields:
                    found_fields.add(path)
                    fields.append(QueryField(path, pivoted, direction, priority))

        limit = settings.DATA_BROWSER_DEFAULT_ROW_LIMIT

        filters = []
        for path__lookup, values in dict(get_args).items():
            for value in values:
                if path__lookup == "limit":
                    try:
                        limit = max(1, int(value))
                    except:  # noqa: E722  input sanitization
                        pass
                if "__" in path__lookup:
                    path, lookup = path__lookup.rsplit("__", 1)
                    filters.append(QueryFilter(path, lookup, value))

        return cls(model_name, fields, filters, limit)

    @property
    def _field_str(self):
        field_strs = []
        for field in self.fields:
            pivot = "&" if field.pivoted else ""
            direction = {ASC: "+", DSC: "-", None: ""}[field.direction]
            priority = str(field.priority) if field.direction else ""
            field_strs.append(f"{pivot}{'__'.join(field.path)}{direction}{priority}")
        return ",".join(field_strs)

    @property
    def _filter_fields(self):
        return [
            ("__".join(filter.path + [filter.lookup]), filter.value)
            for filter in self.filters
        ] + [("limit", self.limit)]

    def get_url(self, media):
        base_url = reverse(
            "data_browser:query",
            kwargs={
                "model_name": self.model_name,
                "fields": self._field_str,
                "media": media,
            },
        )
        params = urllib.parse.urlencode(
            self._filter_fields, quote_via=urllib.parse.quote_plus, doseq=True
        )
        return f"{base_url}?{params}"


class BoundFieldMixin:
    @cached_property
    def path(self):
        return self.orm_bound_field.full_path

    @cached_property
    def pretty_path(self):
        return self.orm_bound_field.pretty_path

    @cached_property
    def path_str(self):
        return self.orm_bound_field.path_str


@dataclass
class BoundFilter(BoundFieldMixin):
    orm_bound_field: Any
    lookup: str
    value: str

    @classmethod
    def bind(cls, orm_bound_field, query_filter):
        return cls(orm_bound_field, query_filter.lookup, query_filter.value)

    def __post_init__(self):
        self.parsed, self.err_message = self.orm_bound_field.parse(
            self.lookup, self.value
        )
        self.is_valid = not self.err_message


@dataclass
class BoundField(BoundFieldMixin):
    orm_bound_field: Any
    pivoted: bool
    direction: Optional[str]
    priority: Optional[int]

    @classmethod
    def bind(cls, orm_bound_field, query_field):
        return cls(
            orm_bound_field,
            query_field.pivoted and orm_bound_field.can_pivot,
            query_field.direction if orm_bound_field.concrete else None,
            query_field.priority if orm_bound_field.concrete else None,
        )


def _orm_fields(fields):
    return [f.orm_bound_field for f in fields]


class BoundQuery:
    def __init__(self, model_name, fields, filters, limit):
        self.model_name = model_name
        self.fields = fields
        self.filters = filters
        self.limit = limit

    @classmethod
    def bind(cls, query, orm_models):
        def get_orm_field(parts):
            model_name = query.model_name
            orm_bound_field = None
            for part in parts:
                if model_name is None:
                    return None
                orm_field = orm_models[model_name].fields.get(part)
                if orm_field is None:
                    return None
                orm_bound_field = orm_field.bind(orm_bound_field)
                model_name = orm_field.rel_name
            if not orm_bound_field.type_:
                return None
            return orm_bound_field

        model_name = query.model_name

        fields = []
        for query_field in query.fields:
            orm_bound_field = get_orm_field(query_field.path)
            if orm_bound_field:
                fields.append(BoundField.bind(orm_bound_field, query_field))

        filters = []
        for query_filter in query.filters:
            orm_bound_field = get_orm_field(query_filter.path)
            if orm_bound_field and orm_bound_field.concrete:
                filters.append(BoundFilter.bind(orm_bound_field, query_filter))

        return cls(model_name, fields, filters, query.limit)

    @cached_property
    def sort_fields(self):
        return sorted((f for f in self.fields if f.direction), key=lambda f: f.priority)

    @cached_property
    def valid_filters(self):
        return [f for f in self.filters if f.is_valid]

    @cached_property
    def col_fields(self):
        return [f for f in self.fields if f.pivoted]

    @cached_property
    def row_fields(self):
        if self.col_fields:
            return [
                f for f in self.fields if f.orm_bound_field.can_pivot and not f.pivoted
            ]
        else:
            return self.fields

    @cached_property
    def data_fields(self):
        if self.col_fields:
            return [f for f in self.fields if not f.orm_bound_field.can_pivot]
        else:
            return []

    @cached_property
    def bound_fields(self):
        return _orm_fields(self.fields)

    @cached_property
    def bound_filters(self):
        return _orm_fields(self.valid_filters)

    @cached_property
    def bound_col_fields(self):
        return _orm_fields(self.col_fields)

    @cached_property
    def bound_row_fields(self):
        return _orm_fields(self.row_fields)

    @cached_property
    def bound_data_fields(self):
        return _orm_fields(self.data_fields)
